# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Scheduler and callbacks (aka observers) to be called during a simulation.

To add a callback `func` to a `Simulation` instance `sim` and have it
called every 100 steps

.. code-block:: python

   sim.add(func, Scheduler(100))

There are two ways to set them up and add them to a simulation:

1. callable classes:

   .. code-block:: python

      sim.add(simulation.WriterThermo(), 100)
      sim.add(simulation.TargetRMSD(5.0))

2. use functions passing optional `*args` and `**kwargs`:

   .. code-block:: python

      sim.add(simulation.writer_thermo, 100)
      sim.add(simulation.target_rmsd, 100, rmsd=5.0)

To differentiate different types of callbacks, we following a naming
convention (for classes and function). If they contain

- target : these callbacks raise a SimulationEnd when it's over
- writer : these callbacks dump useful stuff to file

Of course, general purpose callback can be passed to do whatever.
"""

import sys
import os
import time
import datetime
import logging
import numpy
from atooms.core.utils import rmd, rmf

__all__ = ['SimulationEnd', 'WallTimeLimit', 'SimulationKill',
           'Scheduler', 'write_config', 'write_thermo', 'store',
           'write_trajectory', 'write', 'target', 'target_rmsd',
           'target_steps', 'target_walltime', 'user_stop',
           'target_user_stop', 'Speedometer', 'shell_stop',
           'target_shell_stop', 'target_python_stop', 'stuck']

_log = logging.getLogger(__name__)


# Helper functions

def _sec2time(time_interval):
    """
    Convert a time interval in seconds to (day, hours, minutes,
    seconds) format.
    """
    eta_d = time_interval / (24.0 * 3600)
    eta_h = (eta_d - int(eta_d)) * 24
    eta_m = (eta_h - int(eta_h)) * 60.0
    eta_s = (eta_m - int(eta_m)) * 60.0
    return '%dd:%02dh:%02dm:%02ds' % (eta_d, eta_h, eta_m, eta_s)


# Default exceptions

class SimulationEnd(Exception):
    """Raised when an targeter reaches its target."""

class SimulationKill(Exception):
    """Raised when a simulation is terminated by SIGTERM."""

class WallTimeLimit(Exception):
    """Raised when the wall time limit is reached."""


# Scheduler classes

class Scheduler:

    """
    Schedule observer calls during the simulation.

    This is nothing but a callable that takes a simulation instance
    and returns the next step at which an observer has to to notified.
    """

    def __init__(self, interval=None, calls=None, steps=None,
                 block=None, seconds=None):
        """
        Only one of the arguments can be different from None.

        - `interval`: notify at a fixed steps interval
        - `calls`: fixed number of notification
        - `steps`: list of steps at which the observer will be notified
        - `block`: as steps, but will be called periodically
        - `seconds`: notify every `seconds`
        """
        self.interval = interval
        self.calls = calls
        self.steps = steps
        self.block = block
        self.seconds = seconds

        # Normalize non-positive intervals and n. of calls
        if self.interval is not None and self.interval <= 0:
            self.interval = None
        if self.calls is not None and self.calls <= 0:
            self.calls = None

    def __call__(self, sim):
        """
        Given a simulation instance `sim`, return the next step at which
        the observer will be called.
        """
        # Regular interval
        if self.interval is not None and self.calls is None:
            return (sim.current_step // self.interval + 1) * self.interval

        # Fixed number of calls
        if self.calls is not None:
            interval = max(1, sim.steps // self.calls)
            return (sim.current_step // interval + 1) * interval

        # List of selected steps
        if self.steps is not None:
            inext = sys.maxsize
            for step in self.steps:
                if step > sim.current_step:
                    inext = step
                    break
            return inext

        # Periodic block of steps
        if self.block is not None:
            step_of_last_block = (sim.current_step // self.block[-1]) * self.block[-1]
            inext = sys.maxsize
            for i, step in enumerate(self.block):
                if step > sim.current_step % self.block[-1]:
                    inext = self.block[i] + step_of_last_block
                    break
            return inext

        # TODO: implement this
        if self.seconds is not None:
            pass

        return sys.maxsize


# Writer callbacks
# Callbacks as pure function to distinguish their role we adopt a naming convention:
# if the callback contains write (target) in its __name__ then it is a writer (targeter).

def write_to_ram(sim, trajectory_ram):
    """
    Write configurations to a trajectory in ram.
    """
    trajectory_ram.write(sim.system, sim.current_step)


def write_trajectory(sim, variables=None, precision=None, trajectory=None,
                     trajectory_class=None, fields=None):
    """
    Write trajectory frame from `sim` Simulation instance

    The trajectory format is taken from `sim.trajectory_class` and a
    local instance with that format is used appending the frames on
    successive calls to this function.

    If `trajectory` is a Trajectory instance, it is used instead.
    """
    # TODO: deprecate fields in favor of variables
    if fields is not None:
        variables = fields

    # Clear up everything
    if sim.current_step == 0 and trajectory is None and \
       trajectory_class is None:
        # TODO: bug here with trajectory the file is removed!
        # TODO: folder-based trajectories should ensure that mode='w' clears up the folder
        rmd(sim.output_path)
        rmf(sim.output_path)

    if trajectory is None and trajectory_class is None:
        th = sim.trajectory_class(sim.output_path, 'a')
    elif trajectory_class is not None:
        th = trajectory_class(sim.output_path, 'a')
    else:
        th = trajectory

    # Write stuff
    if hasattr(sim.backend, 'timestep'):
        th.timestep = sim.backend.timestep
    if precision is not None:
        th.precision = precision
    if variables is not None:
        th.variables = variables
    th.write(sim.system, sim.current_step)

    if trajectory is None:
        th.close()


# Deprecated alias
write_config = write_trajectory

def write_thermo(sim, fields=None, fmt=None, precision=6, functions=None):
    """
    Write thermodynamic properties to a file.

    By default, available fields are:
    - steps
    - temperature
    - potential energy per particle
    - kinetic energy per particle
    - total energy
    - pressure
    - rmsd
    - cm velocity norm

    The set of available `fields` can be augmented by passing an extra
    `functions` dictionary.

    The `fmt` dictionary can be used to provide custom formatting
    options for fields.

    The `precision` parameter controls the default precision of
    floating point fields.
    """
    # TODO: deprecate fields
    # By default write minimal info
    if fields is None:
        fields = ['steps',
                  'temperature',
                  'potential energy per particle',
                  'kinetic energy per particle',
                  'total energy per particle',
                  'rmsd',
                  'cm velocity norm']

    # Internal function database.
    # It can be augmented via functions parameter
    _db_func = {
        'steps': lambda x: x.current_step,
        'potential energy per particle': lambda x: x.system.potential_energy(True),
        'kinetic energy per particle': lambda x: x.system.kinetic_energy(True),
        'total energy per particle': lambda x: x.system.total_energy(True, cache=True),
        'temperature': lambda x: x.system.temperature,
        'density': lambda x: x.system.density,
        'pressure': lambda x: x.system.pressure,
        'rmsd': lambda x: x.rmsd,
        'cm velocity': lambda x: x.system.center_of_mass('velocity'),
        'cm velocity norm': lambda x: numpy.sum(x.system.center_of_mass('velocity')**2)**0.5,
    }

    # Update db with extra functions
    if functions is not None:
        _db_func.update(functions)

    # Internal database for formats
    _db_fmt = {}
    for key in _db_func:
        # Default to float formatting
        _db_fmt[key] = '{{:.{precision}{form}}}'.format(precision=precision, form='g')
    # Steps are integer
    _db_fmt['steps'] = '{:d}'

    # Update db with extra formats
    if fmt is not None:
        _db_fmt.update(fmt)

    # Header
    if sim.current_step == 0:
        with open(sim.output_path + '.thermo', 'w') as fh:
            txt = ', '.join(fields)
            fh.write('# columns: {}\n'.format(txt))

    # Line
    with open(sim.output_path + '.thermo', 'a') as fh:
        values = [_db_func[field](sim) for field in fields]
        result = ' '.join([_db_fmt[field].format(value) for value, field in zip(values, fields)])
        fh.write('{}\n'.format(result))


def _setup_callbacks(what):
    """Setup callbacks from `what` list, see `write` for definitions"""

    # Default callbacks that take simulation as first argument
    _callbacks = {
        'steps': lambda x: x.current_step,
        'potential energy per particle': lambda x: x.system.potential_energy(True),
        'kinetic energy per particle': lambda x: x.system.kinetic_energy(True),
        'total energy per particle': lambda x: x.system.total_energy(True, cache=True),
        'potential_energy_per_particle': lambda x: x.system.potential_energy(True),
        'kinetic_energy_per_particle': lambda x: x.system.kinetic_energy(True),
        'total_energy_per_particle': lambda x: x.system.total_energy(True, cache=True),
        'temperature': lambda x: x.system.temperature,
        'density': lambda x: x.system.density,
        'pressure': lambda x: x.system.pressure,
        'rmsd': lambda x: x.rmsd,
        'cm velocity': lambda x: x.system.center_of_mass('velocity'),
        'cm velocity norm': lambda x: numpy.sum(x.system.center_of_mass('velocity')**2)**0.5,
        'cm_velocity': lambda x: x.system.center_of_mass('velocity'),
        'cm_velocity_norm': lambda x: numpy.sum(x.system.center_of_mass('velocity')**2)**0.5,
    }

    names, callbacks = [], []
    for attribute in what:
        if attribute in _callbacks:
            # A predefined callback
            names.append(attribute)
            callbacks.append(_callbacks[attribute])
        elif isinstance(attribute, (list, tuple)):
            # A tuple or list (name, callback)
            assert len(attribute) == 2
            names.append(attribute[0])
            callbacks.append(attribute[1])
        else:
            # Generic simulation attribute
            names.append(attribute)
            callbacks.append('{simulation.' + attribute + '}')

    return names, callbacks


def write(sim, what, suffix=None, path=None):
    """
    Write generic attributes of simulation `sim` to a file.

    `suffix` is a tag appended to `sim.output_path` to define the output
    file path. If `path` is provided, however, it is used instead as
    output file path.

    `what` tells the function what to write and must be a list of either:

    - a string representing a valid property of the Simulation instance `sim`

    - a tuple `(name, callback)` where `name` is a descriptive name of
    the value returned by the `callback`, which is a function that
    takes `sim` as first argument

    - a string from the following list:
      - steps
      - potential_energy_per_particle
      - kinetic_energy_per_particle
      - total_energy_per_particle
      - temperature
      - density
      - pressure
      - rmsd
      - cm_velocity
      - cm_velocity_norm
    """
    # TODO: add formats / precision
    assert suffix is not None or path is not None
    assert not (suffix is not None and path is not None)

    if path is None:
        path = sim.output_path + '.' + suffix

    # Define callbacks
    names, callbacks = _setup_callbacks(what)

    # Extract the requested attribute
    values = []
    for callback in callbacks:
        if callable(callback):
            values.append(callback(sim))
        else:
            # This is a string to be formatted with the simulation variables.
            # Then we get it back with its guessed type.
            values.append(callback.format(simulation=sim))

    # Header
    if sim.current_step == 0:
        with open(path, 'w') as fh:
            fh.write('# columns: {}\n'.format(', '.join(names)))

    # Format output string
    fmt = ('{} ' * len(values)) + '\n'
    with open(path, 'a') as fh:
        fh.write(fmt.format(*values))


def store(sim, what, data=None, db=None):
    """
    Store generic attributes of simulation `sim` in the dictonary `data`.

    `what` tells the function what to write and must be a list of either:

    - a string representing a valid property of the Simulation instance `sim`

    - a tuple `(name, callback)` where `name` is a descriptive name of
      the value returned by the `callback`, which is a function that
      takes `sim` as first argument

    - a string from the following list:

      - `steps`
      - `potential_energy_per_particle`
      - `kinetic_energy_per_particle`
      - `total_energy_per_particle`
      - `temperature`
      - `density`
      - `pressure`
      - `rmsd`
      - `cm_velocity`
      - `cm_velocity_norm`
    """
    if db is not None:
        import warnings
        warnings.warn('db argument to store is deprecated, use data instead', DeprecationWarning)
        data = db

    if data is None:
        if not hasattr(sim, 'data'):
            sim.data = {}
        data = sim.data

    # Define callbacks
    names, callbacks = _setup_callbacks(what)

    # If the dict is empty fill it
    if len(data) == 0:
        for name in names:
            data[name] = []

    # Extract the requested attribute
    for name, callback in zip(names, callbacks):
        from atooms.core.utils import tipify
        if callable(callback):
            data[name].append(callback(sim))
        else:
            # This is a string to be formatted with the simulation variables.
            # Then we get it back with its guessed type.
            val = tipify(callback.format(simulation=sim))
            data[name].append(val)

    return data


# Target callbacks.
# They should return a fractional measure of completion

def target(sim, attribute, value):
    """
    An observer that raises a `SimulationEnd` exception when a given
    target `value` of a property is reached during a simulation. The
    property is `attribute` and is assumed to be an attribute of
    simulation or a function receiving sim as first argument.

    Return: the ratio between current and target values of the attribute.
    """
    if isinstance(attribute, str):
        x = float(getattr(sim, attribute))
    else:
        x = attribute(sim)
    if value > 0:
        frac = x / value
        _log.debug('target {} now at {} [{:d}]'.format(attribute, x, int(frac * 100)))
    if x >= value:
        raise SimulationEnd('reached target {}: {}'.format(attribute, value))
    return frac

def target_rmsd(sim, value):
    """Target the root mean squared displacement."""
    return target(sim, 'rmsd', value)

def target_steps(sim, value):
    """Target the number of steps."""
    if sim.current_step >= value:
        raise SimulationEnd('reached target steps %d' % value)
    return float(sim.current_step) / value

def target_walltime(sim, value):
    """
    Target a value of the elapsed wall time in seconds from the
    beginning of the simulation.

    Useful to self restarting jobs in a queining system with time
    limits.
    """
    wtime_limit = value
    if sim.wall_time() > wtime_limit:
        raise SimulationEnd('target wall time reached')

    t = sim.wall_time()
    dt = wtime_limit - t
    _log.debug('elapsed time %g, reamining time %g', t, dt)


def target_python_stop(sim, condition):
    """
    Stop the simulation if `condition` is True.

    `condition` will interpolate attributes of the passed `sim`
    instance. For instance, the condition

    .. code-block:: python

       {current_step} > 1000 and {rmsd} > 1.0

    will stop the simulation when the step is > 1000 and the rmsd > 1.
    """
    # We do nothing on the first step
    if sim.current_step == 0:
        return
    # Interpolate the command string
    cmd = condition.replace('{', '{0.')
    cmd = cmd.format(sim)
    if eval(cmd):
        raise SimulationEnd('condition "{}" satisfied'.format(condition))

def shell_stop(sim, cmd, exit_code=1):
    """
    Execute the shell command `cmd` and stop the simulation if the
    command returns an exit value equal to `exit_code`.

    `cmd` is actually a format string that may contain references to
    the passed `sim` instance. For instance, a valid command is

    .. code-block:: python

        echo {sim.current_step} {sim.rmsd} >> {sim.output_path}.out

    which will append the step and rmsd to {sim.output_path}.out.
    """
    import subprocess
    # We do nothing on the first step
    if sim.current_step == 0:
        return
    try:
        # Interpolate the command string
        wrap_cmd = cmd.format(sim=sim)
        # Run the shell command
        output = subprocess.check_output(wrap_cmd, shell=True,
                                         stderr=subprocess.STDOUT,
                                         executable="/bin/bash")
        if len(output) > 0:
            _log.info('shell command "{}" returned: {}'.format(cmd, output.strip()))

    except subprocess.CalledProcessError as e:
        # We stop the simulation
        if e.returncode == exit_code:
            raise SimulationEnd('shell command "{}" returned "{}"'.format(cmd, e.output.strip()))
        _log.error('shell command {} failed with output {}'.format(cmd, e.output))
        raise

def user_stop(sim):
    """
    Allows a user to stop the simulation smoothly by touching a STOP
    file in the output root directory.  Currently the file is not
    deleted to allow parallel jobs to all exit.
    """
    # To make it work in parallel we should broadcast and then rm
    # or subclass userstop in classes that use parallel execution
    if sim.output_path is not None:
        if os.path.isdir(sim.output_path):
            dirpath = sim.output_path
        else:
            dirpath = os.path.dirname(sim.output_path)
        if os.path.exists('%s/STOP' % dirpath):
            raise SimulationEnd('user has stopped the simulation')

class stuck:

    """Stop the simulation if the system gets stuck in configuration space"""

    def __init__(self):
        self.estimates = 10
        self._last_value = 0.0
        self._last_step = 0
        self._derivatives = []
        self._values = []
        self._steps = []

    def __call__(self, sim):
        if sim.current_step == 0:
            return
        current = sim.rmsd
        deriv = (current - self._last_value) / (sim.current_step - self._last_step)
        self._derivatives.append(deriv)
        self._values.append(current)
        self._steps.append(sim.current_step)
        if len(self._derivatives) < self.estimates:
            return

        n = self.estimates // 2
        deriv = numpy.mean(self._derivatives[:n])
        # print(sim.current_step, current, (deriv - std) * (sim.current_step - self._last_step))
        # print(numpy.array(self._values[-n:]))
        actual = numpy.array(self._values[-n:])
        estimate = self._values[n] + deriv * (numpy.array(self._steps[-n:]) - self._steps[-n])
        if numpy.all(actual < estimate):
            raise SimulationEnd('system got stuck')

        self._last_value = current
        self._last_step = sim.current_step
        self._derivatives = self._derivatives[-self.estimates:]
        self._values = self._values[-self.estimates:]
        self._steps = self._steps[-self.estimates:]


# Aliases
target_user_stop = user_stop
target_shell_stop = shell_stop


class Speedometer:

    """Display speed of simulation and remaining time to reach target."""

    def __init__(self, log=True, path=None):
        self._init = False
        self._log = log
        self._path = path

    def __str__(self):
        return 'speedometer'

    def __call__(self, sim):
        if self._path is None and sim.output_path is not None:
            self._path = sim.output_path + '.speed.yaml'
        if not self._init:
            # We could store all this in __init__() but this
            # way we allow targeters added to simulation via add()
            for observer in sim._observer:
                c = observer['callback']
                if c is self:
                    continue
                name = getattr(c, '__name__', 'unknown').lower()
                if 'target' in name:
                    self._observer = observer
                    args = observer['args']
                    kwargs = observer['kwargs']
                    self.x_last = c(sim, *args, **kwargs)
                    self.t_last = time.time()
                    self._init = True
                    return

        t_now = time.time()
        args = self._observer['args']
        kwargs = self._observer['kwargs']
        x_now = self._observer['callback'](sim, *args, **kwargs)
        # Get the speed at which the simulation advances
        speed = (x_now - self.x_last) / (t_now - self.t_last)
        # Report fraction of target achieved and ETA
        frac = float(x_now) / 1
        try:
            eta = (1.0 - x_now) / speed
            d_now = datetime.datetime.now()
            d_delta = datetime.timedelta(seconds=eta)
            d_eta = d_now + d_delta
            # self._callback.__name__,
            if self._log:
                _log.info('%2d%% ETA: %s S/T: %.1f T/SP: %.2e',
                          int(frac * 100),
                          d_eta.strftime('%Y-%m-%d %H.%M'),
                          1./sim.wall_time(per_step=True),
                          sim.wall_time(per_step=True, per_particle=True))
            else:
                with open(self._path, 'w') as fh:
                    eta = d_eta.strftime('%Y-%m-%d %H.%M')
                    fh.write(f"""
completed: {int(frac * 100)}%
ETA: {eta}
T/SP: {sim.wall_time(per_step=True, per_particle=True)}
""")
        except ZeroDivisionError:
            print(x_now, self.x_last)
            raise

        self.t_last = t_now
        self.x_last = x_now



"""Store trajectory in memory (it can be huge)."""

import copy
from atooms.system import System, Particle, Cell
from .base import TrajectoryBase


class TrajectoryRam(TrajectoryBase):

    """
    Store trajectory in RAM

    The read_system method of this class conforms with the normal
    Trajectory behavior, i.e. a copy of the system is returned when
    requesting the same frame multiple times.
    """

    def __init__(self, data=None, filename=None, mode='w'):
        super(TrajectoryRam, self).__init__(filename, mode)
        self._system = []
        self._overwrite = True
        if data:
            self._from_data(data)

    def _from_data(self, data):
        """
        Create a system from a dictionary `data`

        The keys of `data` must be valid `what` entries for
        `System.view()`.
        """
        # TODO: accept more general System attributes beyond particle and cell
        # We need the number of particles, we look for it in
        # `particle.position`, which must be present
        from atooms.core.utils import canonicalize
        N = None
        for key in data:
            if canonicalize([key], self.thesaurus)[0] == 'particle.position':
                pos = data[key][0]
                N = pos.shape[0]
                break
        assert N, f'could not guess particle number from {data.keys()}'

        # All items of the dict must have the same len (i.e. the number of frames)
        frames = list({len(data[key]) for key in data})
        assert len(frames) == 1
        frames = int(frames[0])
        for frame in range(frames):
            s = System()
            s.particle = [Particle() for _ in range(N)]
            s.cell = Cell()
            for what in data:
                s.view(what)[...] = data[what][frame]
            self.append(s)

    def write_system(self, system, step):
        if step in self.steps:
            ind = self.steps.index(step)
            self._system[ind].update(system)
        else:
            self._system.append(copy.deepcopy(system))
            # Ensure the system cache is cleared
            # TODO: relax this once order changes are handled in System.dump()
            self._system[-1].dump(clear=True)

    def read_system(self, frame):
        return copy.deepcopy(self._system[frame])

    def __setitem__(self, i, value):
        try:
            step = self.steps[i]
        except IndexError:
            if len(self.steps) > 0:
                step = self.steps[-1]+1
            else:
                step = 0
        self.write(value, step)


class TrajectoryRamView(TrajectoryRam):

    """
    This class deviates from the normal Trajectory behavior in that it
    returns views on the System when calling read_system(), and not
    copies. Thus modifications to the read system object will be
    propagated to the trajectory.
    """

    def read_system(self, frame):
        return self._system[frame]


class TrajectoryRamLight(TrajectoryRam):

    """
    Light-weight version
    """

    def write_system(self, system, step):
        """
        Store copies of the views of the attributes of the System
        components
        """
        # We treat differently particles from the rest
        # Store all instance variables
        data = {}
        data['particle'] = {}
        for key in system.particle[0].__dict__.keys():
            value = system.view('particle.' + key)
            # Restore C order (transpose as (N, ndim))
            # We rely on the underlying storage order to detect this
            if value.flags['F_CONTIGUOUS']:
                value = value.transpose()
            data['particle'][key] = value.copy()

        data['cell'] = {}
        if system.cell is not None:
            for key in system.cell.__dict__.keys():
                data['cell'][key] = system.view('cell.' + key).copy()

        # Insert frame
        if step in self.steps:
            ind = self.steps.index(step)
            self._system[ind] = data
        else:
            self._system.append(data)

    def read_system(self, frame):
        # TODO: optimize by storing a private System
        N = len(self._system[frame]['particle']['species'])
        s = System()
        s.particle = [Particle() for _ in range(N)]
        s.cell = Cell()

        data = self._system[frame]
        particle_attrs = data['particle'].keys()
        cell_attrs = data['cell'].keys()

        for what in particle_attrs:
            for i, p in enumerate(s.particle):
                setattr(p, what, data['particle'][what][i].copy())
        for what in cell_attrs:
            setattr(s.cell, what, data['cell'][what].copy())

        # # Set all default attributes
        # default_particle_attrs = s.particle[0].__dict__.keys()
        # # Note: setting as view can be more efficient only if we use a private system
        # # Otherwise s.view() internally iterates over all particles
        # for what in default_particle_attr:
        #     s.view(f'particle.{what}')[...] = data['particle'][what].copy()
        # for what in cell_attr:
        #     s.view(f'cell.{what}')[...] = data['cell'][what].copy()

        # # Now set custom attributes. These cannot be set as views properly right now.
        # for what in set(stored_particle_attr) - set(default_particle_attr):
        #     for i, p in enumerate(s.particle):
        #         setattr(p, what, data['particle'][what][i].copy())
        return s


# This is maintanined for backward compatibility
TrajectoryRamFull = TrajectoryRamView



#!/usr/bin/env python

import os
import unittest
from atooms.core import utils


class Test(unittest.TestCase):

    def setUp(self):
        self.dirbase = '/tmp/test_core'
        self.dirout = self.dirbase + '/dir'
        utils.mkdir(self.dirout)
        with open(self.dirout + '/1', 'w'):
            pass
        utils.cp(self.dirout + '/1', self.dirout + '/2')
        utils.mkdir(self.dirout + '/3')

    def test_rm(self):
        utils.rmf([self.dirout + '/1', self.dirout + '/2'])
        self.assertFalse(os.path.exists(self.dirout + '/1'))
        self.assertFalse(os.path.exists(self.dirout + '/2'))
        self.assertTrue(os.path.exists(self.dirout + '/3'))

    def test_rm_glob(self):
        utils.rmf(self.dirout + '/*')
        self.assertFalse(os.path.exists(self.dirout + '/1'))
        self.assertFalse(os.path.exists(self.dirout + '/2'))
        self.assertTrue(os.path.exists(self.dirout + '/3'))

    def test_rmd(self):
        utils.rmd(self.dirbase)

    def test_wget(self):
        utils.wget('https://framagit.org/atooms/atooms/-/raw/master/README.md', self.dirbase)
        utils.download('https://framagit.org/atooms/atooms/-/raw/master/README.md', self.dirbase)

    def test_timer(self):
        t = utils.Timer()
        t.start()
        t.stop()
        t.wall_time
        t.cpu_time

    def test_tipify(self):
        from atooms.core.utils import tipify
        self.assertTrue(type(tipify("2.0")) is float)
        self.assertTrue(type(tipify("2")) is int)
        self.assertTrue(type(tipify("t2")) is str)


if __name__ == '__main__':
    unittest.main()



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""PDB format, write-only."""

from .base import TrajectoryBase
from atooms.system.particle import Particle
from atooms.system.cell import Cell
from atooms.system import System


class TrajectoryPDB(TrajectoryBase):

    """PDB format (https://en.wikipedia.org/wiki/Protein_Data_Bank_(file_format))"""

    suffix = 'pdb'

    def __init__(self, filename, mode='r'):
        super(TrajectoryPDB, self).__init__(filename, mode)
        self._file = open(self.filename, self.mode)
        if mode == 'r':
            self._setup_index()

    def read_steps(self):
        pass

    def _setup_index(self):
        self._index_frame = []
        self.steps = []
        fh = self._file
        self._index_header_lines = 0
        header = True
        while True:
            line = fh.tell()
            data = fh.readline()
            # We break if file is over or we found an empty line
            if not data:
                break

            # Count number of header lines
            if header:
                self._index_header_lines += 1

            # Store lines of frame beginnings
            if data.startswith('MODEL'):
                # End of the header
                if header:
                    header = False
                self.steps.append(int(data.split()[-1]))
                self._index_frame.append(line)

    def write_system(self, system, step):
        cfg = ''
        cfg += 'MODEL%9i\n' % step
        fmt = 'CRYST1' + 3*'{:9.3f}' + 3*'{:7.2f}' + ' P 1           1\n'
        cfg += fmt.format(*(list(system.cell.side) + [90, 90, 90]))
        for p in system.particle:
            # If particle has a field property we dump it in the pdb file
            if hasattr(p, 'field'):
                x = float(p.field)
            else:
                x = 1.0
            data = [' '] * 80
            data[0:6] = 'HETATM'
            data[30:38] = '{:8.3f}'.format(p.position[0])
            data[38:46] = '{:8.3f}'.format(p.position[1])
            data[46:54] = '{:8.3f}'.format(p.position[2])
            data[60:66] = str(x)
            data[76:77] = p.species
            cfg += ''.join(data) + '\n'
        cfg += 'ENDMDL\n'
        self._file.write(cfg)

    def read_init(self):
        self._cell = None
        self._file.seek(0)
        for _ in range(self._index_header_lines):
            data = self._file.readline()
            if data.startswith('CRYST1'):
                Lx = float(data[6:15])
                Ly = float(data[15:24])
                Lz = float(data[24:33])
                self._cell = Cell([Lx, Ly, Lz])

    def read_system(self, frame):
        particle = []
        cell = self._cell

        # Read system at frame
        self._file.seek(self._index_frame[frame])
        _ = self._file.readline()
        while True:
            data = self._file.readline()
            if not data or data.startswith('ENDMDL'):
                break
            if data.startswith('CRYST1'):
                Lx = float(data[6:15])
                Ly = float(data[15:24])
                Lz = float(data[24:33])
                cell = Cell([Lx, Ly, Lz])
            elif data.startswith('HETATM') or data.startswith('ATOM'):
                atom = data[0:6]
                serial = data[6:11]
                name = data[12:16]
                altLoc = data[16]
                resName = data[17:20]
                chainID = data[21]
                resSeq = data[22:26]
                iCode = data[26]
                x = float(data[30:38])
                y = float(data[38:46])
                z = float(data[46:54])
                occupancy = data[54:60]
                tempFactor = data[60:66]
                try:
                    element = data[76:77]
                    charge = data[78:80]
                except IndexError:
                    element = name
                particle.append(Particle(position=[x, y, z], species=element))
        system = System()
        system.particle = particle
        system.cell = cell
        return system

    def close(self):
        self._file.close()



"""
Thermodynamic reservoirs to impose temperature, pressure, or
particle numbers in a simulation.

When added to a `System` object, they will determine the statistical
ensemble of a simulation:

.. code-block:: python

   system = System()
   system.thermostat = Thermostat(temperature=1.0)
   system.barostat = Barostat(pressure=1.0)

Note: in an actual simulation backend, these reservoirs will have
additional degrees of freedom, e.g. s and pi in a Nose-like
thermostat.
"""

import numpy


class Thermostat(object):

    """Thermostat to control the temperature during a simulation."""

    def __init__(self, temperature, name='', mass=1.0, coordinate=1.0,
                 momentum=1.0, relaxation_time=1.0,
                 collision_period=None):
        self.name = name
        self.temperature = temperature
        self.coordinate = numpy.asarray(coordinate)
        self.momentum = numpy.asarray(momentum)
        self.mass = numpy.asarray(mass)
        self.relaxation_time = relaxation_time
        if collision_period is not None:
            self.relaxation_time = collision_period

    @property
    def collision_period(self):
        return self.relaxation_time


class Barostat(object):

    """Barostat to control the pressure during a simulation."""

    def __init__(self, pressure, name='', mass=1.0, coordinate=1.0,
                 momentum=1.0, relaxation_time=1.0):
        self.name = name
        self.pressure = pressure
        self.coordinate = numpy.asarray(coordinate)
        self.momentum = numpy.asarray(momentum)
        self.mass = numpy.asarray(mass)
        self.relaxation_time = relaxation_time


class Reservoir(object):

    """Reservoir to control the number of particles during a simulation."""

    def __init__(self, chemical_potential, name='', mass=1.0):
        self.name = name
        self.chemical_potential = chemical_potential
        self.mass = mass



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Base and total interaction class.

Actual interaction backends should implement this interface or
subclass Interaction by implementing the `compute()` method and
specifying the variables passed to it using the Interaction.variables
dictionary.
"""

import numpy


class InteractionBase(object):

    """Base interaction class"""

    def __init__(self):
        self.variables = {'position': 'particle.position'}
        """
        A dictionary of variables needed to compute the interaction

        The keys must match the variable names in the
        `Interaction.compute()` interface, the fields must be
        canonicalizable variables accepted by `System.dump()`.

        It is possible to specify the required data type using the
        optional colon syntax `<property>[:<dtype>]`. The `dtype` must a
        valid identifier for `numpy` array creation.
        """
        # TODO: order is not a good variable, we should expand the syntax using [:order]
        self.order = 'F'  # deprecated
        self.observable = ['energy', 'forces', 'virial', 'stress', 'hessian']
        for observable in self.observable:
            setattr(self, observable, None)

    def __add__(self, other):
        total = Interaction()
        for attr in self.observable:
            if getattr(self, attr, None) is None or getattr(other, attr, None) is None:
                continue
            elif getattr(self, attr) is not None and getattr(other, attr) is not None:
                # Store the sum of the properties in the total interaction
                setattr(total, attr, getattr(self, attr) + getattr(other, attr))
            else:
                raise ValueError('attribute {} not set in {} or {}'.format(attr,
                                                                           self,
                                                                           other))
        return total

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def compute(self, observable, position):
        """Compute an `observable` from this interaction

        Subclasses must adapt the interface so as to match the keys specified by
        the `Interaction.variables` dict.

        :param observable: the observable to compute ('energy', 'forces',
            'stress', 'hessian')
        :param position: input position array
        """
        # Sanity checks
        assert observable in self.observable, \
            'unsupported observable {}'.format(observable)

        # Zeroing observables
        ndim, N = position.shape
        if observable == 'energy':
            self.energy = 0.0
        elif observable == 'forces' or observable is None:
            self.energy = 0.0
            self.virial = 0.0
            self.forces = numpy.zeros_like(position)
        elif observable == 'stress':
            self.energy = 0.0
            self.virial = 0.0
            self.forces = numpy.zeros_like(position)
            self.stress = numpy.zeros(ndim, ndim)
        elif observable == 'hessian':
            self.hessian = numpy.zeros((ndim, N, ndim, N))


class Interaction(InteractionBase):

    """Total interaction comprising multiple interaction terms"""

    def __init__(self, *terms):
        InteractionBase.__init__(self)
        self.variables = {}
        self.term = []
        for term in terms:
            self.add(term)

    def add(self, term):
        """Add an `Interaction` term """
        self.term.append(term)
        self.variables.update(term.variables)

    def compute(self, observable, **kwargs):
        """Compute an `observable` from all the terms of this interaction

        :param observable: the observable to compute ('energy', 'forces',
            'stress', 'hessian')
        :param **kwargs: keyword arguments to be passed to the interaction term
        """
        if len(self.term) == 0:
            return

        for term in self.term:
            # Extract the relevant variables for this term
            term_kwargs = {}
            for key in term.variables:
                term_kwargs[key] = kwargs[key]
            term.compute(observable, **term_kwargs)

        # Sum all interaction terms
        total = sum(self.term)
        for attr in self.observable:
            setattr(self, attr, getattr(total, attr))


class InteractionWall(InteractionBase):

    """Interaction between particles and walls"""

    def __init__(self, wall, potential):
        InteractionBase.__init__(self)
        self.potential = potential
        if not isinstance(wall, (list, tuple)):
            self.wall = [wall]
        else:
            self.wall = wall
        self.variables = {'pos': 'particle.position',
                          'ids': 'particle.species:int32'}

    def compute(self, observable, pos, ids):
        """Compute an `observable` from this `InteractionWall` instance

        :param observable: the observable to compute ('energy', 'forces', None)
        :param pos: position array
        :param ids: species array
        """
        assert observable in ['forces', 'energy', None]
        if self.forces is None:
            self.forces = numpy.empty_like(pos, order='F')
        self.energy, self.virial, self.forces[:, :] = 0.0, 0.0, 0.0
        N = len(ids)
        # These arrays could be cached
        u, w, h = numpy.ndarray(N), numpy.ndarray(N), numpy.ndarray(N)
        for wall in self.wall:
            rij = wall.distance(pos)
            rsq = numpy.sum(rij**2, axis=0)
            self.potential(rsq, u, w, h)
            self.energy += numpy.sum(u)
            self.forces += w[:] * rij[:, :]


class InteractionField(InteractionBase):

    """Interaction between particles and an external field"""

    def __init__(self, field, variables=None):
        InteractionBase.__init__(self)
        self.field = field
        self.variables = variables
        if variables is None:
            self.variables = {'pos': 'particle.position'}

    def compute(self, observable, **variables):
        """Compute the `observable` from this `InteractionField` instance

        :param observable: the observable to compute ('energy', 'forces', None)
        :param **variables: keyword arguments to compute the interaction; they
            must include the particle position as `pos` keyword argument and
            additional variables to be passed to the `field` function
        """
        assert 'pos' in variables
        assert observable in ['forces', 'energy', None]
        if self.forces is None:
            self.forces = numpy.empty_like(variables['pos'], order='F')
        u, grad, _ = self.field(**variables)
        self.energy = numpy.sum(u)
        self.virial = 0.0
        self.forces = - grad[:, :]



import numpy
import f2py_jit
from .helpers import _merge_source


class NeighborList:

    def __init__(self, rcut, neighbors='neighbor_list.f90',
                 helpers='helpers.f90', inline=True, full=False,
                 max_neighbors=-1, method='fixed_cutoff'):
        self.rcut = numpy.array(rcut)
        self.neighbors = None
        self.distances = None
        self.number_neighbors = None
        self.full = full
        self.max_neighbors = max_neighbors
        self.method = method
        self._module_path = None

        # Cut off check
        if method == "sann":
            self.rcut = numpy.max(self.rcut)

        # Gather f90 sources into a single one
        source = _merge_source(helpers, neighbors)

        # Inline subroutine
        if inline:
            from f2py_jit.finline import inline_source
            source = inline_source(source, ignore='compute')  # avoid reinlining forces!

        # Build the module with f2py and store module uid
        # (better not store the module itself, else we cannot deepcopy)
        # extra_args = '--opt="-O3 -ffast-math"'
        extra_args = '--opt="-fbounds-check"'
        self._uid = f2py_jit.build_module(source, extra_args=extra_args)

        # TODO: add parallel and debug flags?
        # This was in atooms-models
        # args, opt_args = '', '-O3 -ffree-form -ffree-line-length-none'
        # if debug:
        #     opt_args += ' -pg -fbounds-check'
        # if parallel:
        #     opt_args += ' -fopenmp'
        #     args += ' -lgomp'
        # extra_args = '--opt="{}" {}'.format(opt_args, args)

    def _setup(self, npart, nneigh):
        """Allocate or reallocate arrays for neighbor list"""
        if self.neighbors is None or self.neighbors.shape[1] != npart or self.neighbors.shape[0] < nneigh:
            self.neighbors = numpy.ndarray(shape=(nneigh, npart), order='F', dtype=numpy.int32)
            self.distances = numpy.ndarray(shape=(nneigh, npart), order='F', dtype=float)
        if self.number_neighbors is None or len(self.number_neighbors) != npart:
            self.number_neighbors = numpy.ndarray(npart, order='F', dtype=numpy.int32)

    def compute(self, box, pos, ids):
        # Setup
        f90 = f2py_jit.import_module(self._uid)

        # Setup arrays
        # Estimate max number of neighbors based on average density
        # We take the largest cut off distance
        npart = pos.shape[1]
        rho = npart / box.prod()
        if self.max_neighbors > 0:
            nneigh = self.max_neighbors
        else:
            nneigh = int(4.0 / 3.0 * 3.1415 * rho * numpy.max(self.rcut)**3 * 1.50)
        self._setup(npart, nneigh)

        # Compute neighbors list
        #
        # If the f90 code returns an error, the arrays are reallocated
        # based on the largest number of neighbors returned by the f90
        # routine
        if self.method == 'fixed_cutoff' and self.full:
            func = f90.neighbor_list.compute_full
        if self.method == 'fixed_cutoff' and not self.full:
            func = f90.neighbor_list.compute
        if self.method == 'sann':
            func = f90.neighbor_list.compute_sann

        error = func(box, pos, ids, self.rcut, self.neighbors, self.number_neighbors, self.distances)
        if error:
            self._setup(npart, max(self.number_neighbors))
            error = func(box, pos, ids, self.rcut, self.neighbors, self.number_neighbors, self.distances)
            assert not error, "something wrong with neighbor_list"



#!/usr/bin/env python

import unittest
import numpy

from atooms.core.utils import mkdir
from atooms.trajectory import Unfolded
from atooms.trajectory import TrajectoryXYZ, TrajectorySimpleXYZ, TrajectoryRUMD, TrajectoryEXYZ


class TestXYZ(unittest.TestCase):

    Trajectory = TrajectoryXYZ

    def setUp(self):
        mkdir('/tmp/test_xyz')
        self.finp = '/tmp/test_xyz/pbc.xyz'
        with open(self.finp, 'w') as fh:
            fh.write("""\
2
cell:6.0,6.0,6.0
A 1.0 -1.0 0.0
A 2.9 -2.9 0.0
2
cell:6.0,6.0,6.0
A 1.1 -1.1 0.0
A -2.9 -2.9 0.0
2
cell:6.0,6.0,6.0
A 1.2 -1.2 0.0
A -2.9 2.9 0.0
2
cell:6.0,6.0,6.0
A 1.3 -1.3 0.0
A -2.8 2.8 0.0
""")
        # Test metadata recognition
        self.finp_meta = '/tmp/test_xyz/meta.xyz'
        with open(self.finp_meta, 'w') as fh:
            fh.write("""\
2
metafmt:space,comma columns:id,x,y,z mass:1.0,2.0 step:1 cell:5.0,5.0,5.0
A 1.0 -1.0 0.0
B 2.9 -2.9 0.0
""")
        # Test 4-dim file
        self.finp_4d = '/tmp/test_xyz/4d.xyz'
        with open(self.finp_4d, 'w') as fh:
            fh.write("""\
2
metafmt:space,comma columns:id,pos ndim:4 mass:1.0,2.0 step:1
A 1.0 -1.0 0.0 1.0
B 2.9 -2.9 0.0 2.0
2
metafmt:space,comma columns:id,pos mass:1.0,2.0 step:1 cell:5.0,5.0,5.0,5.0
A 1.0 -1.0 0.0 1.0
B 2.9 -2.9 0.0 2.0
2
metafmt:space,comma columns:id,pos mass:1.0,2.0 step:1 cell:5.0,5.0,5.0
A 1.0 -1.0 0.0 1.0
B 2.9 -2.9 0.0 2.0
""")

    def test_xyz_4d(self):
        with self.Trajectory(self.finp_4d) as t:
            meta = t._read_comment(0)
            self.assertEqual(meta['ndim'], 4)
            self.assertEqual(len(t[0].particle[0].position), 4)
            self.assertEqual(t[0].particle[1].position[3], 2.0)
            # Test to check we grab ndim from the n. of cell coordinates
            meta = t._read_comment(1)
            self.assertEqual(meta['ndim'], 4)
            self.assertEqual(len(t[1].particle[0].position), 4)
            self.assertEqual(t[1].particle[1].position[3], 2.0)
            # This last test shows that when cell has 3 coords, we are back to ndim=3
            meta = t._read_comment(2)
            self.assertEqual(meta['ndim'], 3)
            self.assertEqual(len(t[2].particle[0].position), 3)

    def test_xyz_field(self):
        # Test fields
        finp = '/tmp/test_xyz/field.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
1
columns:field step:1
1.0
""")
        with self.Trajectory(finp) as t:
            x = t[0].particle[0].field
            self.assertEqual(x, 1.0)

    def test_xyz_meta_with_spaces(self):
        # Test fields
        finp = '/tmp/test_xyz/spaces.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
1
columns:id step:1 cell:    1.0,   1.0,  1.0
1
""")
        with self.Trajectory(finp) as t:
            x = t[0].cell.side
            self.assertEqual(x[0], 1.0)
            self.assertEqual(x[1], 1.0)
            self.assertEqual(x[2], 1.0)

    def test_xyz_particle(self):
        self.finp = '/tmp/test_xyz/particle.xyz'
        import atooms.system
        system = atooms.system.System(N=108)
        with self.Trajectory(self.finp, 'w') as th:
            th.write(system)
        with self.Trajectory(self.finp, 'r') as th:
            system = th[0]
            self.assertEqual(system.distinct_species, ['A'])
                                
    def test_xyz_meta(self):
        with self.Trajectory(self.finp_meta) as t:
            meta = t._read_comment(0)
            self.assertEqual(t.steps, [1])
            self.assertEqual(meta['mass'], [1, 2])
            self.assertEqual(t[0].particle[0].mass, 1.0)
            self.assertEqual(t[0].particle[1].mass, 2.0)

    def test_xyz_indexed(self):
        r_ref = [[1., -1., 0.], [2.9, -2.9, 0.]]
        with self.Trajectory(self.finp) as t:
            self.assertEqual(t.steps, [1, 2, 3, 4])
            self.assertEqual(r_ref[0], list(t[0].particle[0].position))
            self.assertEqual(r_ref[1], list(t[0].particle[1].position))

    def test_xyz_indexed_unfolded(self):
        t1 = self.Trajectory(self.finp)
        t = Unfolded(t1)
        t[0]
        t[1]
        self.assertEqual(list(t[2].particle[1].position), [3.1, -3.1, 0.0])
        self.assertEqual(list(t1[2].particle[1].position), [-2.9, 2.9, 0.0])
        t1.close()
        t.close()

    def test_xyz_indexed_unfolded_skip(self):
        """Test that unfolded trajectories can skip samples"""
        # Here we read all samples
        t1 = self.Trajectory(self.finp)
        t = Unfolded(t1)
        t[0]
        t[1]
        t[2]
        s = t[3]
        t1.close()
        t.close()

        # Here we skip one sample
        t1 = self.Trajectory(self.finp)
        t1 = Unfolded(t1)
        t1[0]
        s1 = t1[3]
        t1.close()

        self.assertEqual(list(s.particle[0].position), list(s1.particle[0].position))
        self.assertEqual(list(s.particle[1].position), list(s1.particle[1].position))

    def test_xyz_unfolded(self):
        """Test reading unfolded after reading normal"""
        with self.Trajectory(self.finp) as t1:
            for s in t1:
                pass
            with Unfolded(t1) as t:
                for s in t:
                    pass

    def test_xyz_with(self):
        r_ref = [[1., -1., 0.], [2.9, -2.9, 0.]]
        with self.Trajectory(self.finp) as t:
            self.assertEqual(r_ref[0], list(t[0].particle[0].position))
            self.assertEqual(r_ref[1], list(t[0].particle[1].position))

        with self.Trajectory(self.finp + '.out', 'w') as to:
            with self.Trajectory(self.finp) as t:
                # This is necessary because timestep is not copied automatically
                # and if the class tries to read it, it won't find
                to.timestep = t.timestep
                to.write(t[0], 0)

        with self.Trajectory(self.finp + '.out') as to:
            self.assertEqual(r_ref[0], list(to[0].particle[0].position))

    def test_xyz_iter(self):
        with self.Trajectory(self.finp) as t:
            for s in t:
                s.particle
            t[-1]

    def test_xyz_to_ram(self):
        # TODO: move this test somewhere else
        from atooms.trajectory import TrajectoryRam
        with self.Trajectory(self.finp) as t:
            with TrajectoryRam() as tram:
                for i, system in enumerate(t):
                    tram[i] = system
                    # Write a second time, this will overwrite
                    tram[i] = system
                    self.assertEqual(tram[i].particle[0].position[0],
                                     system.particle[0].position[0])

    def test_xyz_ids(self):
        # Test ids ordering
        finp = '/tmp/test_xyz/meta.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
2
metafmt:space,comma columns:id,x,y,z step:1 cell:5.0,5.0,5.0
B 1.0 -1.0 0.0
A 2.9 -2.9 0.0
2
metafmt:space,comma columns:id,x,y,z step:2 cell:5.0,5.0,5.0
C 1.0 -1.0 0.0
B 2.9 -2.9 0.0
""")
        with self.Trajectory(finp) as th:
            self.assertEqual(th[0].particle[0].species, 'B')
            self.assertEqual(th[0].particle[1].species, 'A')
            self.assertEqual(th[1].particle[0].species, 'C')
            self.assertEqual(th[1].particle[1].species, 'B')

    def test_xyz_mass(self):
        finp = '/tmp/test_xyz/meta.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
3
metafmt:space,comma columns:id,x,y,z,radius mass:1.0,2.0,3.0 step:1 cell:5.0,5.0,5.0
B 1.0 -1.0 0.0 0.5
A 2.9 -2.9 0.0 0.6
C 2.9 -2.9 0.0 0.7
3
metafmt:space,comma columns:id,x,y,z,radius mass:2.0,3.0 step:1 cell:5.0,5.0,5.0
C 1.0 -1.0 0.0 0.5
B 2.9 -2.9 0.0 0.6
B 2.9 -2.9 0.0 0.7
""")
        with self.Trajectory(finp) as th:
            self.assertEqual(th[0].particle[0].radius, 0.5)
            self.assertEqual(th[0].particle[1].radius, 0.6)
            self.assertEqual(th[0].particle[2].radius, 0.7)
            self.assertEqual(th[0].particle[0].mass, 2.0)
            self.assertEqual(th[0].particle[1].mass, 1.0)
            self.assertEqual(th[0].particle[2].mass, 3.0)
            self.assertEqual(th[1].particle[0].mass, 3.0)
            self.assertEqual(th[1].particle[1].mass, 2.0)
            self.assertEqual(th[1].particle[2].mass, 2.0)

        # This is ok
        with open(finp, 'w') as fh:
            fh.write("""\
3
columns:id,x,y,z mass:2 step:1
B 1.0 -1.0 0.0
A 2.9 -2.9 0.0
C 2.9 -2.9 0.0
""")
        with self.Trajectory(finp) as th:
            self.assertEqual(th[0].particle[0].mass, 2.0)
            self.assertEqual(th[0].particle[1].mass, 2.0)
            self.assertEqual(th[0].particle[2].mass, 2.0)

        # This is not ok, mass / species mismatch
        with open(finp, 'w') as fh:
            fh.write("""\
3
columns:id,x,y,z mass:1,2 step:1
B 1.0 -1.0 0.0
A 2.9 -2.9 0.0
C 2.9 -2.9 0.0
""")
        with self.Trajectory(finp) as th:
            with self.assertRaises(ValueError):
                th[0]

    def test_xyz_columns(self):
        finp = '/tmp/test_xyz/columns.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
1
columns:name,pos,vel step:1
A 1.0 -1.0 0.0 1.0 2.0 3.0
""")
        with self.Trajectory(finp) as th:
            self.assertEqual(th._read_comment(0)['columns'], ['name', 'pos', 'vel'])
            self.assertEqual(list(th[0].particle[0].velocity), [1.0, 2.0, 3.0])

    def test_xyz_cell(self):
        """Test that fluctuating cell side is read correctly."""
        finp = '/tmp/test_xyz/cell.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
1
step:0 cell:1.0,1.0,1.0
A 1.0 -1.0 0.0
1
step:1 cell:2.0,1.0,1.0
A 1.0 -1.0 0.0
1
step:2 cell:1.0,1.0,1.0
A 1.0 -1.0 0.0
1
step:3 cell:3.0,1.0,1.0
A 1.0 -1.0 0.0
""")
        with TrajectoryXYZ(finp) as th:
            self.assertEqual([s.cell.side[0] for s in th], [1.0, 2.0, 1.0, 3.0])

    def test_xyz_any(self):
        finp = '/tmp/test_xyz/any.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
1
step:0 cell:1.0,1.0,1.0 columns:species,position,ox,oy,oz
A 0.0 0.0 0.0 1.0 1.0 1.0
""")
        with TrajectoryXYZ(finp) as th:
            th[0].particle[0].oz

    def test_xyz_position_unfolded(self):
        """Test that unfolded positions are parsed, if present, and propagated to folded ones"""
        finp = '/tmp/test_xyz/unfolded.xyz'
        with open(finp, 'w') as fh:
            fh.write("""\
1
step:0 cell:2.0,2.0,2.0 columns:species,position_unfolded
A 1.1 11.1 21.1
""")
        with TrajectoryXYZ(finp) as th:
            s = th[0]
            self.assertAlmostEqual(s.particle[0].position_unfolded[0], 1.1)
            self.assertAlmostEqual(s.particle[0].position[0], -0.9)
            self.assertAlmostEqual(s.particle[0].position[1], -0.9)
            self.assertAlmostEqual(s.dump('particle.position_unfolded')[0, 0], 1.1)
            self.assertAlmostEqual(s.dump('position_unfolded')[0, 0], 1.1)

    def tearDown(self):
        from atooms.core.utils import rmd
        rmd('/tmp/test_xyz')


from atooms.trajectory.xyz import TrajectoryNeighbors

class TestNeighbors(unittest.TestCase):

    Trajectory = TrajectoryNeighbors

    def setUp(self):
        mkdir('/tmp/test_xyz')

    def test_neighbors_consume(self):
        with open('/tmp/test_xyz/neighbors.xyz', 'w') as fh:
            fh.write("""\
4
step:1 columns:neighbors*
2 4
1 3
2
1
""")
        with TrajectoryNeighbors('/tmp/test_xyz/neighbors.xyz', offset=0) as th:
            s = th[0]
            self.assertEqual(list(s.particle[0].neighbors), [2, 4])
            self.assertEqual(list(s.particle[1].neighbors), [1, 3])
            self.assertEqual(list(s.particle[2].neighbors), [2])
            self.assertEqual(list(s.particle[3].neighbors), [1])

    def test_neighbors_comma(self):
        with open('/tmp/test_xyz/neighbors.xyz', 'w') as fh:
            fh.write("""\
4
step:1 columns:neighbors timestep:0.001
2,4
1,3
2
1
""")
        with TrajectoryNeighbors('/tmp/test_xyz/neighbors.xyz', offset=0) as th:
            s = th[0]
            self.assertEqual(th.timestep, 0.001)
            self.assertEqual(list(s.particle[0].neighbors), [2, 4])
            self.assertEqual(list(s.particle[1].neighbors), [1, 3])
            self.assertEqual(list(s.particle[2].neighbors), [2])
            self.assertEqual(list(s.particle[3].neighbors), [1])

    def tearDown(self):
        from atooms.core.utils import rmd
        rmd('/tmp/test_xyz')


class TestSimpleXYZ(TestXYZ):

    Trajectory = TrajectorySimpleXYZ

    def test_xyz_field(self):
        pass

    def test_xyz_4d(self):
        pass

    def test_xyz_meta(self):
        pass

    def test_xyz_meta_with_spaces(self):
        pass

    def test_xyz_mass(self):
        pass

    def test_xyz_columns(self):
        pass

    def test_xyz_cell(self):
        pass


class TestRumd(unittest.TestCase):

    Trajectory = TrajectoryRUMD

    def setUp(self):
        mkdir('/tmp/test_xyz')
        super(TestRumd, self).setUp()
        ioformat_1 = """\
2
ioformat=1 dt=0.005000000 boxLengths=6.000000000,6.000000000,6.000000000 numTypes=2 Nose-Hoover-Ps=-0.154678583 Barostat-Pv=0.000000000 mass=1.000000000,1.000000000 columns=type,x,y,z,imx,imy,imz,vx,vy,vz,fx,fy,fz,pe,vir
0  1.0 -1.0 0.0 0 0 1 -0.955896854 -2.176721811 0.771060944 14.875996590 -28.476327896 -15.786120415 -5.331668218 22.538120270
1  2.9 -2.9 0.0 0 0 -1 -0.717318892 -0.734904408 0.904034972 -28.532371521 13.714955330 0.387423307 -7.276362737 11.813765526
"""

        with open('/tmp/test_xyz/rumd.xyz', 'w') as fh:
            fh.write(ioformat_1)
            self.finp = fh.name

    def test_xyz_particle(self):
        self.finp = '/tmp/test_xyz/particle.xyz'
        import atooms.trajectory
        import atooms.system
        system = atooms.system.System(N=108)
        with atooms.trajectory.TrajectoryRUMD(self.finp, 'w') as th:
            th.write(system)
        with atooms.trajectory.TrajectoryRUMD(self.finp, 'r') as th:
            system = th[0]
            self.assertEqual(system.distinct_species, ['0'])
                                            
    def test_read_write(self):
        with TrajectoryRUMD(self.finp) as th:
            with TrajectoryRUMD(self.finp + 'out', 'w') as th_out:
                self.assertEqual(th.timestep, 0.005)
                self.assertEqual(list(th[0].cell.side), [6.0, 6.0, 6.0])
                s = th[0]
                imxs, imys, imzs = s.dump('imx'), s.dump('imy'), s.dump('imz')
                self.assertEqual(list(imxs), [0, 0])
                self.assertEqual(list(imzs), [1, -1])
                th_out.timestep = th.timestep
                for i, system in enumerate(th):
                    th_out.write(system, th.steps[i])
        with TrajectoryRUMD(self.finp + 'out') as th:
            self.assertEqual(th.timestep, 0.005)
            self.assertEqual(list(th[0].cell.side), [6.0, 6.0, 6.0])


class TestUtils(unittest.TestCase):

    Trajectory = TrajectoryXYZ

    def setUp(self):
        mkdir('/tmp/test_xyz')
        with open('/tmp/test_xyz/1.xyz', 'w') as fh:
            fh.write("""\
1
metafmt:space,comma columns:id,x,y,z step:1 cell:5.0,5.0,5.0
B 1.0 -1.0 0.0
1
metafmt:space,comma columns:id,x,y,z step:2 cell:5.0,5.0,5.0
B 2.9 -2.9 0.0
1
metafmt:space,comma columns:id,x,y,z step:4 cell:5.0,5.0,5.0
B 2.9 -2.9 0.0
""")
        with open('/tmp/test_xyz/2.xyz', 'w') as fh:
            fh.write("""\
1
metafmt:space,comma columns:id,x,y,z step:2 cell:5.0,5.0,5.0
B 1.0 -1.0 0.0
1
metafmt:space,comma columns:id,x,y,z step:3 cell:5.0,5.0,5.0
B 2.9 -2.9 0.0
1
metafmt:space,comma columns:id,x,y,z step:4 cell:5.0,5.0,5.0
B 2.9 -2.9 0.0
1
metafmt:space,comma columns:id,x,y,z step:5 cell:5.0,5.0,5.0
B 2.9 -2.9 0.0
""")

    def test_paste(self):
        import atooms.trajectory as trj
        t1 = trj.TrajectoryXYZ('/tmp/test_xyz/1.xyz')
        t2 = trj.TrajectoryXYZ('/tmp/test_xyz/2.xyz')
        steps = []
        for step, s1, s2 in trj.utils.paste(t1, t2):
            steps.append(step)
        self.assertEqual(steps, [2, 4])
        t1.close()
        t2.close()

    def tearDown(self):
        from atooms.core.utils import rmd
        rmd('/tmp/test_xyz')


class TestEXYZ(unittest.TestCase):

    Trajectory = TrajectoryEXYZ

    def setUp(self):
        mkdir('/tmp/test_xyz')
        self.finp = '/tmp/test_xyz/pbc.xyz'
        self.fout = '/tmp/test_xyz/out.xyz'
        with open(self.finp, 'w') as fh:
            fh.write("""\
2
Lattice="5.44 0.0 0.0 0.0 5.44 0.0 0.0 0.0 5.44" Properties=species:S:1:pos:R:3 Time=0.0
A 1.0 -1.0 0.0
B 2.9 -2.9 0.0
2
Lattice="5.44 0.0 0.0 0.0 5.44 0.0 0.0 0.0 5.44" Properties=species:S:1:pos:R:3 Time=0.0
A 1.1 -1.1 0.0
C -2.9 -2.9 0.0
""")

    def test_xyz_any(self):
        self.finp = '/tmp/test_xyz/pbc.xyz'
        with open(self.finp, 'w') as fh:
            fh.write("""\
1
Properties=species:S:1:pos:R:3:orientation:R:3 Lattice="5.44 0.0 0.0 0.0 5.44 0.0 0.0 0.0 5.44"
A 0.0 0.0 0.0 1.0 1.0 1.0
""")
        with self.Trajectory(self.finp) as th:
            th[0].particle[0].orientation

    def test_xyz_meta(self):
        with self.Trajectory(self.finp) as th:
            meta = th._read_comment(0)
            system_0 = th[0]
            system_1 = th[1]
            self.assertEqual(th.steps, [1, 2])
            self.assertEqual(th[0].particle[0].species, 'A')
            self.assertEqual(th[0].particle[1].species, 'B')
            self.assertEqual(th[1].particle[0].species, 'A')
            self.assertEqual(th[1].particle[1].species, 'C')
            self.assertAlmostEqual(th[0].particle[0].position[0], 1.0)
            self.assertAlmostEqual(th[0].particle[1].position[1], -2.9)
            self.assertAlmostEqual(th[1].particle[0].position[0], 1.1)
            self.assertAlmostEqual(th[1].particle[1].position[1], -2.9)

            # self.assertEqual(meta['mass'], [1, 2])
        with self.Trajectory(self.fout, 'w') as th:
            th.write(system_0, 1)
            th.write(system_1, 2)

        with self.Trajectory(self.fout) as th:
            meta = th._read_comment(0)
            system_0 = th[0]
            system_1 = th[1]
            self.assertEqual(th.steps, [1, 2])
            self.assertEqual(th[0].particle[0].species, 'A')
            self.assertEqual(th[0].particle[1].species, 'B')
            self.assertEqual(th[1].particle[0].species, 'A')
            self.assertEqual(th[1].particle[1].species, 'C')
            self.assertAlmostEqual(th[0].particle[0].position[0], 1.0)
            self.assertAlmostEqual(th[0].particle[1].position[1], -2.9)
            self.assertAlmostEqual(th[1].particle[0].position[0], 1.1)
            self.assertAlmostEqual(th[1].particle[1].position[1], -2.9)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import unittest

from atooms.optimization import Optimization
import atooms.backends.lammps
from atooms.backends.lammps import EnergyMinimization

SKIP = not atooms.backends.lammps.installed()

class Test(unittest.TestCase):

    def setUp(self):
        import os
        if SKIP:
            self.skipTest('missing LAMMPS')
        self.input_file = os.path.join(os.path.dirname(__file__),
                                       '../data/lj_N256_rho1.0.xyz')

    def test_min(self):
        cmd = """
        pair_style      lj/cut 2.5
        pair_modify     shift yes
        pair_coeff      1 1 1.0 1.0 2.5
        """
        bck = EnergyMinimization(self.input_file, cmd)
        bck.verbose = False
        opt = Optimization(bck)
        opt.tolerance = 1e-4
        e = bck.system.potential_energy(per_particle=True)
        opt.run()
        e = bck.system.potential_energy(per_particle=True)
        w = bck.system.force_norm_square(per_particle=True)
        self.assertLess(e, -6.8)
        self.assertLess(w, 1e-4)


if __name__ == '__main__':
    unittest.main()



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

import numpy
import re
from .base import TrajectoryBase
from .utils import gopen
from atooms.core.utils import tipify
from atooms.system.particle import Particle
from atooms.system.cell import Cell
from atooms.system import System


class TrajectorySimpleXYZ(TrajectoryBase):

    """
    Simple implementation of the xyz layout (https://en.wikipedia.org/wiki/XYZ_file_format)

    It uses a memory light-weight indexed access.
    """

    suffix = 'xyz'

    def __init__(self, filename, mode='r'):
        super(TrajectorySimpleXYZ, self).__init__(filename, mode)
        self._cell = None
        # Trajectory file handle
        self._file = gopen(self.filename, self.mode)
        if self.mode == 'r':
            # Internal index of lines to seek and tell.
            # We may delay setup, moving to read_init() assuming
            # self.steps becomes a property
            self._setup_index()

    def read_steps(self):
        steps = []
        for frame in range(len(self._index_frame)):
            meta = self._read_comment(frame)
            try:
                steps.append(meta['step'])
            except KeyError:
                # If no step info is found, we add steps sequentially
                steps.append(frame+1)
        return steps

    def _setup_index(self):
        """Sample indexing via tell / seek"""
        self._index_frame = []
        self._index_header = []
        self._index_cell = None
        self._file.seek(0)
        while True:
            line = self._file.tell()
            data = self._file.readline().strip()

            # We break if file is over or we found an empty line
            if not data:
                break

            # The first line contains the number of particles
            # If something goes wrong, this could be the last line
            # with the cell side (Lx,Ly,Lz) and we parse it some other way
            try:
                npart = int(data)
                self._index_header.append(line)
            except ValueError:
                self._index_cell = line
                break

            # Skip npart+1 lines
            _ = self._file.readline()
            self._index_frame.append(self._file.tell())
            for _ in range(npart):
                self._file.readline()

    def _read_comment(self, frame):
        """Internal xyz method to get header metadata from comment line of given `frame`.

        We assume metadata format is a space-separated sequence of
        comma separated entries such as:

        columns:id,x,y,z step:10
        columns=id,x,y,z step=10
        """
        # Go to line and skip Npart info
        self._file.seek(self._index_header[frame])
        npart = int(self._file.readline())
        data = self._file.readline()

        # Fill metadata dictionary
        meta = {}
        meta['npart'] = npart
        for e in data.split():
            s = re.search(r'(\S+)\W*[=:]\W*(\S+)', e)
            if s is not None:
                tag, data = s.group(1), s.group(2)
                # Remove dangling commas
                data = data.strip(',')
                # If there are commas, this is a list, else a scalar.
                # We convert the string to appropriate types
                if ',' in data:
                    meta[tag] = [tipify(_) for _ in data.split(',')]
                else:
                    meta[tag] = tipify(data)
        return meta

    def read_init(self):
        # Grab cell from the end of file if it is there
        try:
            side = self._read_comment(0)['cell']
            self._cell = Cell(side)
        except KeyError:
            self._cell = self._parse_cell()

    def _parse_cell(self):
        """Internal emergency method to grab the cell."""
        cell = None
        if self._index_cell:
            self._file.seek(self._index_cell)
            side = numpy.fromstring(self._file.readline(), sep=' ')
            cell = Cell(side)
        return cell

    def read_system(self, frame):
        meta = self._read_comment(frame)
        self._file.seek(self._index_frame[frame])

        # Read particles
        particle = []
        for _ in range(meta['npart']):
            data = self._file.readline().strip().split()
            species = data[0]
            r = numpy.array(data[1:4], dtype=float)
            particle.append(Particle(species=species, position=r))

        # Read cell
        try:
            side = meta['cell']
            self._cell = Cell(side)
        except KeyError:
            pass

        return System(particle, self._cell)

    def _comment_header(self, step, system):
        fmt = "step:%d columns:id,x,y,z" % step
        if system.cell is not None:
            fmt += " cell:" + ','.join(['%s' % x for x in system.cell.side])
        return fmt

    def write_system(self, system, step):
        self._file.write("%s\n" % len(system.particle))
        self._file.write(self._comment_header(step, system) + '\n')
        ndim = len(system.particle[0].position)
        fmt = "%s" + ndim*" %14.6f" + "\n"
        for p in system.particle:
            self._file.write(fmt % ((p.species,) + tuple(p.position)))

    def close(self):
        self._file.close()



from __future__ import absolute_import

import warnings
import numpy as np

# With absolute import (default in python 3) this does not clash
import gsd
import gsd.hoomd

from atooms.system.particle import Particle
from atooms.system.cell import Cell
from atooms.system import System
from atooms.trajectory.base import TrajectoryBase


class TrajectoryGSD(TrajectoryBase):

    """
    Glotzer group's binary GSD format for HOOMD (https://glotzerlab.engin.umich.edu/hoomd-blue/)
    """
    suffix = 'gsd'

    def __init__(self, filename, mode='r', fields=None):
        super(TrajectoryGSD, self).__init__(filename, mode)
        self.variables = ['particle.species', 'particle.position']
        if fields is not None:
            self.variables = fields
            warnings.warn('fields is deprecated, use variables instead', FutureWarning)

        # self.mode can be 'w' or 'r', but gsd is a binary format, so it only accepts 'wb' or 'rb'.
        try:
            version = gsd.version.version[0]
        except AttributeError:
            version = gsd.version.__version__[0]

        if int(version) == 2:
            file_mode = self.mode + "b"
        else:
            file_mode = self.mode
            warnings.warn('gsd version >2 is not supported')

        # Trajectory file handle
        self._file = gsd.hoomd.open(name=self.filename, mode=file_mode)
        # When reading, we must define the steps.
        if self.mode == 'r':
            self.steps = [snap.configuration.step for snap in self._file]

    def read_system(self, frame):
        """ returns System instance. """
        snap = self._file[frame]
        ndim = snap.configuration.dimensions

        # Convert typeid from [0, 0, 1, ...] to ['A', 'A', 'B', ...] when snap.particles.types = ['A', 'B']
        distinct_species = snap.particles.types
        distinct_typeids = list(range(len(distinct_species)))
        typeid_to_species = {}
        for i in distinct_typeids:
            typeid_to_species[i] = distinct_species[i]

        box = snap.configuration.box[:ndim]    # atooms does not handle sheared boxes.
        cell = Cell(side=box)

        N = snap.particles.position.shape[0]
        particles = []
        for i in range(N):
            p = Particle(
                mass=snap.particles.mass[i],
                species=typeid_to_species[snap.particles.typeid[i]],
                position=snap.particles.position[i, :ndim],
                velocity=snap.particles.velocity[i, :ndim],
                radius=snap.particles.diameter[i] / 2
            )
            particles.append(p)

        return System(particle=particles, cell=cell)

    def write_system(self, system, step):
        """ Writes to the file handle self.trajectory."""
        variables = self.variables
        data = {what: system.dump(what) for what in ['particle.position', 'particle.velocity',
                                                     'particle.species', 'particle.mass', 'particle.radius']}
        box = system.cell.side
        N = len(system.particle)
        distinct_species = system.distinct_species

        # Convert species from ['A', 'A', 'B', ...] to [0, 0, 1, ...] when distinct_species = ['A', 'B']
        species_to_typeid = {}
        typeid = 0
        for species in distinct_species:
            species_to_typeid[species] = typeid
            typeid += 1

        pos = data['particle.position']
        vel = data['particle.velocity']
        species = data['particle.species']    # This is 'A', 'A', 'B', etc when distinct_species = ['A', 'B']
        mass = data['particle.mass']
        radius = data['particle.radius']
        convert_to_typeid = np.vectorize(lambda species: species_to_typeid[species])
        typeid = convert_to_typeid(species).astype(int)  # This is 0, 0, 1, etc when distinct_species = ['A', 'B']

        # TODO: make this work with gsd 3
        snap = gsd.hoomd.Snapshot()
        snap.configuration.box = [box[0], box[1], box[2], 0, 0, 0]  # Assume all strains 0.
        snap.configuration.step = step
        snap.particles.types = distinct_species
        snap.particles.N = N

        snap.particles.position = pos   # atooms.system and gsd both save positions from -L/2 to L/2.
        snap.particles.typeid = typeid
        if 'particle.velocity' in variables:
            snap.particles.velocity = vel
        if 'particle.mass' in variables:
            snap.particles.mass = mass
        if 'particle.diameter' in variables:
            snap.particles.diameter = 2 * radius

        self._file.append(snap)



#!/usr/bin/env python

import os
import shutil
import unittest
import numpy

from atooms.core.utils import mkdir, rmd
from atooms.trajectory.folder import TrajectoryFolder


class TestXYZ(unittest.TestCase):

    def setUp(self):
        self.dirname = '/tmp/test_folder'
        rmd(self.dirname)
        mkdir(self.dirname)
        for i in range(10, 13):
            fname = os.path.join(self.dirname, 'step_%d' % i)
            with open(fname, 'w') as fh:
                fh.write("""\
2
step:%d
A 1.0 -1.0 0.0
A 2.9 -2.9 0.0
""" % i)

    def test_xyz_indexed(self):
        r_ref = [[1., -1., 0.], [2.9, -2.9, 0.]]
        t = TrajectoryFolder(self.dirname, step_pattern='step_(\d*)')
        self.assertEqual(t.steps, [10, 11, 12])

    def tearDown(self):
        if os.path.exists(self.dirname):
            shutil.rmtree(self.dirname)


if __name__ == '__main__':
    unittest.main()



import os
import numpy
from numpy import copysign
from atooms.system import System
from atooms.trajectory.hdf5 import TrajectoryHDF5


class TrajectoryModes(TrajectoryHDF5):

    def read_system(self, frame):
        system = System()
        s = self._file["trajectory/realtime/sampleindex"].keys()[frame]
        system.eigenvalues = self._file["trajectory/normalmodes/eigenvalues/%s" % s][:]
        system.eigenfreq = numpy.array([copysign(abs(x)**0.5, x) for x in system.eigenvalues])
        mode_idx = self._file["trajectory/normalmodes/eigenvectors/index/%s" % s][:]
        eigv = {}
        for idx in mode_idx:
            # Modes are F-indexed
            omega = system.eigenfreq[idx-1]
            vect = self._file["trajectory/normalmodes/eigenvectors/vector/%s_mode_%05d" % (s, idx)][:]
            eigv[omega] = vect
        system.eigenvectors = eigv

        # Add positions
        parent = os.path.splitext(self._file.filename)[0]
        step = self.steps[frame]
        with TrajectoryHDF5(parent) as th:
            # The step should be there
            parent_frame = th.steps.index(step)
            system.particle = th[parent_frame].particle
            system.cell = th[parent_frame].cell
        return system

# def test(f):
#     with TrajectoryModes('/tmp/config.h5.sad.inm') as inm:
#         print inm[0]
#         print inm[0].particle
#         print inm[0].eigenvalues[:4] # array, full
#         print inm[0].eigenvectors.values()[:4] # dict, what if we have two identical eigenvalues? or list of arrays but we need the list of w

# test('')


if __name__ == '__main__':
    import sys
    with TrajectoryModes(sys.argv[1]) as inm:
        for i in range(len(inm.steps)):
            w = inm[i].eigenfreq[1]
            # dict, what if we have two identical eigenvalues? or list of arrays but we need the list of w
            print(w, inm[i].particle[0].position, inm[i].eigenvectors[w][0])



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""HDF5 trajectory format."""

import numpy
import h5py
import logging
import warnings

from .base import TrajectoryBase
from atooms.core import ndim
from atooms.system import System
from atooms.system.particle import Particle
from atooms.system.cell import Cell
from atooms.system.interaction import Interaction

_log = logging.getLogger(__name__)


class _CutOff(object):

    def __init__(self, scheme, radius):
        self.scheme = scheme
        self.radius = radius
        self.rcutsq = radius**2
        self.radius_mid = radius
        self.radius_mid_sq = radius**2


class _PairPotential(object):

    interacting_bodies = 2

    def __init__(self, func, params, species, cutoff=None,
                 npoints=20000):
        self.func = func
        self.params = params
        self.species = species
        self.cutoff = cutoff
        self.npoints = npoints


#  Helper functions and classes

class _SafeFile(h5py.File):
    # TODO: decorate hdf5 class so that error messages contain the path of the offending file
    def create_group_safe(self, group):
        # TODO: recursively create all h5 groups that do not exist?
        # TODO: redefine create_group unless this is a serious performace issue?
        if group not in self:
            self.create_group(group)

def _get_cached_list_h5(fh, h5g, data):
    """ Replace |data| with read data in group |h5g| only if data is None """
    if data is None:
        try:
            data = [entry[0] for entry in fh[h5g].values()]
        except:
            data = []
    return data

def _write_datasets(fh, group, datasets):
    """Write several data sets stored in datasets dict in group of fh"""
    for name, dataset in datasets.items():
        fh[group + name] = dataset

def add_interaction_hdf5(finp, ff):
    """Add interaction to hdf5 file"""
    import os
    import glob

    pid = os.getpid()
    f_ref = '/tmp/cnv_%s.h5' % pid
    # TODO: we can cache a ref file if ff is the same
    os.system('system.x -n 2 -f %s %s 1>/dev/null 2>/dev/null' % (ff, f_ref))
    ref = h5py.File(f_ref, 'r')
    fout = finp + '.bak'

    # Add interaction
    os.system('/bin/cp %s %s' % (finp, fout))
    h5 = h5py.File(fout, 'r+')
    # Make sure interaction does not exist
    try:
        del h5['initialstate/interaction']
    except:
        pass
    h5.copy(ref['initialstate/interaction'], 'initialstate/interaction')

    # Final cleanup
    h5.close()
    os.remove(finp)
    os.rename(fout, finp)
    ref.close()
    for f in glob.glob(f_ref + '*'):
        os.remove(f)


class TrajectoryHDF5(TrajectoryBase):

    """In-house trajectory layout in HDF5 format."""

    suffix = 'h5'

    def __init__(self, filename, mode='r+'):
        super(TrajectoryHDF5, self).__init__(filename, mode)

        self.general_info = {}
        self._grandcanonical = False
        self._system = None
        # TODO: move this to write mode
        self.variables = ['particle.position', 'particle.velocity']
        if self.mode == 'r' or self.mode == 'r+':
            self._file = h5py.File(self.filename, mode)
            # gather general info on file
            for entry in self._file['/']:
                if type(self._file[entry]) == h5py.Dataset:
                    self.general_info[entry] = self._file[entry]

        elif self.mode == 'w' or self.mode == 'r+' or self.mode == "w-":
            self._file = _SafeFile(self.filename, self.mode)

        else:
            raise ValueError('Specify mode (r/w) for file %s (invalid: %s)' % (self.filename, self.mode))

    def read_steps(self):
        return [d[0] for d in self._file['trajectory/realtime/stepindex'].values()]

    def read_len(self):
        return len(self._file['trajectory/realtime/stepindex'])

    def close(self):
        try:
            self._file.close()
        except ValueError:
            _log.error('file %s already closed', self.filename)
            raise

    def read_timestep(self):
        try:
            return self._file['trajectory/realtime/timestep'][0]
        except:
            _log.warning('no time step in file %s, set to 1', self.filename)
            return 1.0

    def write_timestep(self, value):
        self._file.create_group_safe('/trajectory')
        self._file.create_group_safe('/trajectory/realtime')
        self._file['trajectory/realtime/timestep'] = [value]

    def read_block_size(self):
        try:
            return self._file['trajectory/realtime/block_period'][0]
        except:
            return None

    def write_block_size(self, value):
        self._file.create_group_safe('/trajectory')
        self._file.create_group_safe('/trajectory/realtime')
        self._file['trajectory/realtime/block_period'] = [value]

    def write_init(self, system):
        from atooms.system.particle import distinct_species
        self._file.create_group_safe('/initialstate')
        self._file['DIMENSIONS'] = [3]
        self._file['NAME_SYS'] = [b'Unknown']
        self._file['VERSION_TRJ'] = [b'1.2']
        self._file['VERSION_MD'] = [b'X.X.X']

        # Particles
        group = '/initialstate/particle/'
        if system.particle is not None:
            self._file.create_group_safe(group)
            particle = system.particle
            species = distinct_species(particle)
            particle_h5 = {'number_of_species': [len(species)],
                           'number_of_particles': [len(particle)],
                           'identity': [species.index(p.species)+1 for p in particle],
                           'element': [str(p.species) for p in particle],
                           'mass': [p.mass for p in particle],
                           'radius': [p.radius for p in particle],
                           'position': [p.position for p in particle],
                           'velocity': [p.velocity for p in particle],
                           }
            _write_datasets(self._file, group, particle_h5)

        # Cell
        group = '/initialstate/cell/'
        if system.cell is not None:
            self._file.create_group_safe(group)
            self._file[group + 'sidebox'] = system.cell.side

        # Thermostat
        group = '/initialstate/thermostat/'
        if system.thermostat is not None:
            self._file.create_group_safe(group)
            self._file[group + 'temperature'] = system.thermostat.temperature
            self._file[group + 'type'] = system.thermostat.name
            self._file[group + 'mass'] = system.thermostat.mass
            self._file[group + 'collision_period'] = system.thermostat.collision_period

        # Interaction
        if system.interaction is not None:
            if type(system.interaction) is list:
                raise TypeError('cannot handle more than one interaction')
            self.write_interaction([system.interaction])

    def write_interaction(self, interaction):
        rgr = '/initialstate/interaction/'
        self._file.create_group_safe('/initialstate/')
        # If the group exisist we delete it. This does not actual clear space in h5 file.
        # We could do it on a dataset basis via require_dataset, or visit the group and delete everything.
        try:
            del self._file['/initialstate/interaction/']
        except:
            pass
        self._file.create_group_safe('/initialstate/interaction/')
        self._file[rgr + 'number_of_interactions'] = [len(interaction)]
        for i, term in enumerate(interaction):
            igr = rgr + '/interaction_%d/' % (i+1)
            # Numbering is fortran style for backward compatibility
            self._file.create_group_safe(igr)
            self._file[igr + 'interaction_type'] = [term.name.encode()]
            self._file[igr + 'number_of_potentials'] = [len(term.potential)]
            for j, phi in enumerate(term.potential):
                pgr = igr + '/potential_%d/' % (j+1)
                self._file.create_group_safe(pgr)
                self._file[pgr + 'potential'] = [str(phi).encode()]
                self._file[pgr + 'interacting_bodies'] = [phi.interacting_bodies]
                self._file[pgr + 'interacting_species'] = phi.species
                self._file[pgr + 'parameters_number'] = [len(phi.params)]
                self._file[pgr + 'parameters_name'] = [_ for _ in sorted(phi.params.keys())]
                self._file[pgr + 'parameters'] = [phi.params[k] for k in sorted(phi.params.keys())]
                self._file[pgr + 'cutoff_scheme'] = [phi.cutoff.scheme.encode()]
                self._file[pgr + 'cutoff_radius'] = [phi.cutoff.radius]
                self._file[pgr + 'lookup_points'] = [phi.npoints]

    def write_system(self, system, step):
        variables = self.variables

        self._file.create_group_safe('/trajectory')
        self._file.create_group_safe('/trajectory/realtime')
        self._file.create_group_safe('/trajectory/realtime/stepindex')
        self._file.create_group_safe('/trajectory/realtime/sampleindex')

        frame = len(self.steps) + 1
        csample = '/sample_%7.7i' % frame

        try:
            self._file['/trajectory/realtime/stepindex' + csample] = [step]
            self._file['/trajectory/realtime/sampleindex' + csample] = [frame]
        except RuntimeError:
            _log.error('error when writing step %s sample %s to file %s', step, frame, self.filename)
            raise

        if system.particle is not None:
            self._file.create_group_safe('/trajectory/particle')
            if 'particle.position' in variables:
                self._file.create_group_safe('/trajectory/particle/position')
                self._file['/trajectory/particle/position' + csample] = [p.position for p in system.particle]
            if 'particle.velocity' in variables:
                self._file['/trajectory/particle/velocity' + csample] = [p.velocity for p in system.particle]
                self._file.create_group_safe('/trajectory/particle/velocity')
            if 'particle.radius' in variables:
                self._file.create_group_safe('/trajectory/particle/radius')
                self._file['/trajectory/particle/radius' + csample] = [p.radius for p in system.particle]
            if 'particle.species' in variables:
                self._file.create_group_safe('/trajectory/particle/species')
                data = ['%-3s' % p.species for p in system.particle]
                self._file['/trajectory/particle/species' + csample] = [_.encode() for _ in data]
                from atooms.system.particle import distinct_species
                ids = distinct_species(system.particle)
                self._file['/trajectory/particle/ids' + csample] = [1+ids.index(p.species) for p in system.particle]

        if system.cell is not None:
            self._file.create_group_safe('/trajectory/cell')
            self._file.create_group_safe('/trajectory/cell/sidebox')
            self._file['/trajectory/cell/sidebox' + csample] = system.cell.side

    def read_init(self):
        # read particles
        group = self._file['/initialstate/particle']
        n = self._file['/initialstate/particle/number_of_particles'][0]
        rad = None
        for entry in group:
            # TODO: refactor this
            if entry == 'element':
                spe = group[entry][:]
            if entry == 'mass':
                mas = group[entry][:]
            if entry == 'position':
                pos = group[entry][:]
            if entry == 'velocity':
                vel = group[entry][:]
            if entry == 'radius':
                rad = group[entry][:]
        if rad is not None:
            particle = [Particle(species=spe[i].decode().strip(), mass=mas[i],
                                 position=pos[i, :], velocity=vel[i, :],
                                 radius=rad[i]) for i in range(n)]
        else:
            particle = [Particle(species=spe[i].decode().strip(), mass=mas[i],
                                 position=pos[i, :], velocity=vel[i, :])
                        for i in range(n)]

        # read cell
        group = self._file['/initialstate/cell']
        for entry in group:
            if entry == 'sidebox':
                sidebox = group[entry][:]
        cell = Cell(sidebox)

        # read interaction
        interaction = self.read_interaction()

        # build system
        self._system = System(particle, cell, interaction)

        return self._system

    def read_interaction(self):
        # read interaction terms
        if 'interaction' not in self._file['/initialstate']:
            return None

        n = self._file['/initialstate/interaction/number_of_interactions'][0]
        if n > 1:
            warnings.warn('can only read one interaction term')

        i = 0
        g = '/initialstate/interaction/interaction_%d/' % (i+1)
        np = self._file[g + 'number_of_potentials'][0]
        name = self._file[g + 'interaction_type'][0].decode()
        potentials = []
        for j in range(np):
            sg = self._file[g + 'potential_%d/' % (j+1)]
            # params = {k:v for k, v in zip(sg['parameters_name'][:], sg['parameters'][:])}
            # make it compatible with 2.6
            params = {}
            for k, v in zip(sg['parameters_name'][:], sg['parameters'][:]):
                params[k] = v
            p = _PairPotential(sg['potential'][0].decode(), params,
                               sg['interacting_species'][:],
                               _CutOff(sg['cutoff_scheme'][0].decode(),
                                       sg['cutoff_radius'][0]))
            if 'lookup_points' in sg:
                p.npoints = sg['lookup_points']

            potentials.append(p)
        interaction = Interaction()
        interaction.potential = potentials
        interaction.name = name
        return interaction

    def read_system(self, frame):
        # TODO: refactor reading particle variables
        # TODO: removing variables in read_system() is pointless, it should be done before
        # We must increase frame by 1 if we iterate over frames with len().
        # This is some convention to be fixed once and for all
        # TODO: read cell on the fly NPT
        keys = list(self._file['/trajectory/realtime/stepindex'].keys())
        csample = '/' + keys[frame]
        # read particles
        group = self._file['/trajectory/particle']
        pos = group['position' + csample][:]

        if 'position_unfolded' in group:
            # fix for unfolded positions that were not written at the first step
            # should be fixed once and for all in md.x
            if frame == 0:
                pos_unf = self._file['/initialstate/particle/position'][:]
            else:
                pos_unf = group['position_unfolded' + csample][:]
            if 'particle.position_unfolded' not in self.variables:
                self.variables.append('particle.position_unfolded')

        if 'velocity' in group and len(group['velocity']) > 0:
            vel = group['velocity' + csample][:]
        else:
            vel = numpy.zeros((len(pos), ndim))
            if 'particle.velocity' not in self.variables:
                self.variables.remove('particle.velocity')

        # Dynamic properties
        p = []
        for r, v in zip(pos, vel):
            p.append(Particle(position=numpy.array(r, dtype='float64'),
                              velocity=numpy.array(v, dtype='float64')))

        # Static properties
        # TODO: optimize, this takes quite some additional time, almost x2
        for pi, po in zip(p, self._system.particle):
            pi.mass = po.mass
            pi.species = po.species
            pi.radius = po.radius

        # Try update radii. This must be done after setting defaults.
        try:
            r = group['radius' + csample][:]
            for i, pi in enumerate(p):
                pi.radius = r[i]
            if 'particle.radius' not in self.variables:
                self.variables.append('particle.radius')
        except KeyError:
            if 'particle.radius' in self.variables:
                self.variables.remove('particle.radius')

        # Try update species. This must be done after setting defaults.
        try:
            spe = group['species' + csample][:]
            for i, pi in enumerate(p):
                pi.species = spe[i].decode().strip()
            if 'particle.species' not in self.variables:
                self.variables.append('particle.species')
        except KeyError:
            if 'particle.species' in self.variables:
                self.variables.remove('particle.species')

        # Read cell
        group = self._file['/trajectory/cell']
        side = group['sidebox' + csample][:]
        # This fixes an issue with some hdf5 trajectories that stored
        # cell as (1,3) array
        if len(side.shape) == 2:
            side = side[0]
        self._system.cell.side = side

        # Read also interaction.
        has_int = True
        try:
            group = self._file['/trajectory/interaction']
        except:
            has_int = False

        if has_int:
            self._system.interaction.total_energy = group['energy' + csample][0]
            self._system.interaction.total_virial = group['virial' + csample][0]
            self._system.interaction.total_stress = group['stress' + csample][:]

        return System(p, self._system.cell, self._system.interaction)



import os
import unittest
from atooms.trajectory import Trajectory
from atooms.system import System, Particle, Cell
from atooms.backends.f90 import Interaction, NeighborList, VerletList


class Test(unittest.TestCase):

    def setUp(self):
        self.fileinp = os.path.join(os.path.dirname(__file__), '../data/lj_N256_rho1.0.xyz')
        self.trajectory = Trajectory(self.fileinp)
        self.model = {
            "potential": [
                {
                    "type": "lennard_jones",
                    "parameters": {"epsilon": [[1.0]], "sigma": [[1.0]]}
                }
            ],
            "cutoff": [
                {
                    "type": "cut_shift",
                    "parameters": {"rcut": [[2.5]]}
                }
            ]
        }

    # TODO: restore when molecular terms are added
    # def test_harmonic_bending(self):
    #     import numpy
    #     from math import pi
    #     angle = 30
    #     model = {
    #         "potential": [
    #             {
    #                 "type": "harmonic_bending",
    #                 "parameters": {"theta0": [[[angle / 360 * 2*pi]]], "k": [[[1.0]]]}
    #             }
    #         ],
    #         "cutoff": [
    #             {
    #                 "type": "cut",
    #                 "parameters": {"rcut": [[2.5]]}
    #             }
    #         ]
    #     }
    #     # x, y = numpy.cos(angle / 360 * 2*pi), numpy.sin(angle / 360 * 2*pi)
    #     # particles = [Particle(position=[0.0, 0.0, 0.0], species=1),
    #     #              Particle(position=[1.0, 0.0, 0.0], species=1),
    #     #              Particle(position=[x, y, 0.0], species=1)]
    #     # cell = Cell([10., 10., 10.])
    #     # system = System(particles, cell)
    #     # system.interaction = Interaction(model, inline=False)
    #     # print(system.potential_energy())
    #     # #self.assertAlmostEqual(system.potential_energy(), 0.0)
    #     # self.assertLess(sum(system.interaction.forces[:, 0]**2), 1e-10)

    #     angle = 60
    #     dx = 0.1
    #     model["potential"][0]["parameters"]["theta0"] = [[[angle / 360 * 2*pi]]]
    #     x, y = numpy.cos(angle / 360 * 2*pi), numpy.sin(angle / 360 * 2*pi)
    #     particles = [Particle(position=[-x+dx, y, 0.0], species=1),
    #                  Particle(position=[0.0, 0.0, 0.0], species=1),
    #                  Particle(position=[x-dx, y, 0.0], species=1)]
    #     cell = Cell([10., 10., 10.])
    #     system = System(particles, cell)
    #     system.interaction = Interaction(model)
    #     system.potential_energy()
    #     self.assertAlmostEqual(system.interaction.forces[0, 0], -system.interaction.forces[0, 2])
    #     self.assertAlmostEqual(system.interaction.forces[0, 1], 0.0)
    #     ref = [[0.2, 0.0, 0.0], [0.9, 0.0, 0.0], [x+0.1, y, 0.0]]

    #     # for i in range(3):
    #     #     dx = 1e-5
    #     #     system.particle[0].position[i] = ref[0][i]
    #     #     u0 = system.potential_energy()
    #     #     system.particle[0].position[i] += dx/2
    #     #     f0 = system.interaction.forces[i, 0]
    #     #     system.particle[0].position[i] += dx/2
    #     #     u1 = system.potential_energy()
    #     #     f = - (u1 - u0) / dx
    #     #     self.assertAlmostEqual(f, f0, places=3)

    #     for i, p in enumerate(system.particle):
    #         p.position[:] = ref[i]

    #     dx = 1e-5
    #     for i in range(3):
    #         system.particle[i].position[0] = ref[i][0]
    #         u0 = system.potential_energy()
    #         system.particle[i].position[0] += dx/2
    #         f0 = system.interaction.forces[0, i]
    #         system.particle[i].position[0] += dx/2
    #         u1 = system.potential_energy()
    #         f = - (u1 - u0) / dx
    #         #self.assertAlmostEqual(f, f0, places=3)
    #         print(f, f0)
    #     return

    #     angle = 60.0
    #     model = {
    #         "potential": [
    #             {
    #                 "type": "harmonic_bending",
    #                 "parameters": {"theta0": [[[angle / 360 * 2*pi]]], "k": [[[1.0]]]}
    #             }
    #         ],
    #         "cutoff": [
    #             {
    #                 "type": "cut",
    #                 "parameters": {"rcut": [[1.01]]}
    #             }
    #         ]
    #     }
    #     x, y = numpy.cos(angle / 360 * 2*pi), numpy.sin(angle / 360 * 2*pi)
    #     particles = [Particle(position=[0.0, 0.0, 0.0], species=1),
    #                  Particle(position=[1.0, 0.0, 0.0], species=1),
    #                  Particle(position=[x, y, 0.0], species=1),
    #                  Particle(position=[-x, y, 0.0], species=1),
    #                  ]
    #     cell = Cell([10., 10., 10.])
    #     system = System(particles, cell)
    #     system.interaction = Interaction(model, inline=True)
    #     # TODO: bug in inline, rjk becomes zero because j is replaced twice with k
    #     # I fixed the offending vars
    #     # ! inline: distance
    #     # ! local: []
    #     # ! dummy: ['j', 'k', 'pos', 'rjk']
    #     # ! vars: ['i', 'j', 'pos', 'rij']
    #     # rjk = pos(:,k) - pos(:,k)
    #     self.assertAlmostEqual(system.potential_energy(), 0.0)

    def test_debug(self):
        from f2py_jit.finline import inline_source
        # The problem is that we first replace i with j, and then k with j (being the
        # second argument. One way to solve the problem is to ensure all local variables
        # are prefixed with _ or something.
        src = """

  pure subroutine distance(i,j,pos,rij)
    integer, intent(in) :: i, j
    double precision, intent(in)    :: pos(:,:)
    double precision, intent(inout) :: rij(:)
    rij = pos(:,i) - pos(:,j)
  end subroutine distance

  subroutine forces()
    double  precision :: pos(:,:)
    integer                         :: i, j, k, jj, isp, jsp, ksp
    double precision                :: rjk(size(pos,1))
    !call distance(jj,k,pos,rjk)  ! OK
    call distance(j,k,pos,rjk)  ! FAILS
end subroutine forces
"""
        inline_source(src)

    def test_collinear(self):
        particles = [Particle(position=[0.0, 0.0, 0.0], species=1),
                     Particle(position=[1.0, 0.0, 0.0], species=1),
                     Particle(position=[2.0, 0.0, 0.0], species=1)]
        cell = Cell([10., 10., 10.])
        system = System(particles, cell)
        system.interaction = Interaction(self.model)
        self.assertAlmostEqual(system.potential_energy(), -0.01257276409199999)

    def test_energy(self):
        from atooms.trajectory.decorators import change_species
        system = self.trajectory[0]
        system.interaction = Interaction(self.model)
        system = change_species(system, 'F')
        system.compute_interaction('energy')
        self.assertAlmostEqual(system.potential_energy(per_particle=True, cache=True), -3.8079776291909284)
        system.compute_interaction('energies')
        self.assertAlmostEqual(system.potential_energy(cache=True), sum(system.interaction.energies) / 2)

    def test_species_layout_cast(self):
        from atooms.trajectory.decorators import change_species
        system = self.trajectory[0]
        system.interaction = Interaction(self.model)
        # Tolerate string species, they will be cast as int
        for p in system.particle:
            p.species = '1'
        self.assertAlmostEqual(system.potential_energy(per_particle=True), -3.8079776291909284)

    def test_species_layout_C(self):
        system = self.trajectory[0]
        system.interaction = Interaction(self.model)
        for p in system.particle:
            p.species = '0'
        self.assertAlmostEqual(system.potential_energy(per_particle=True), -3.8079776291909284)
        system = self.trajectory[0]
        system.interaction = Interaction(self.model, neighbor_list=VerletList())
        for p in system.particle:
            p.species = '0'
        self.assertAlmostEqual(system.potential_energy(per_particle=True), -3.8079776291909284)

    def test_species_layout_A(self):
        # This will fail
        system = self.trajectory[0]
        system.interaction = Interaction(self.model)
        with self.assertRaises(ValueError):
            system.potential_energy(per_particle=True)

    def test_derivatives(self):
        from atooms.trajectory.decorators import change_species
        system = self.trajectory[0]
        system.interaction = Interaction(self.model)
        system = change_species(system, 'F')
        system.compute_interaction("forces")
        self.assertAlmostEqual(system.interaction.forces[0, 0], 65.3083639139375)
        system.compute_interaction('gradw')
        self.assertAlmostEqual(system.interaction.gradw[0, 0], -71505.20734345658)
        system.compute_interaction('hessian')
        self.assertAlmostEqual(system.interaction.hessian[0, 0, 0, 0], 963.3891474196369)

    def test_multiple(self):
        """Test with multiple potentials: compute two LJ interactions with epsilons that sum up to 1"""
        from atooms.trajectory.decorators import change_species
        model = {
            "cutoff": [
                {"type": "cut_shift", "parameters": {"rcut": [[2.5]]}},
                {"type": "cut_shift", "parameters": {"rcut": [[2.5]]}}
            ],
            "potential": [
                {"type": "lennard_jones", "parameters": {"epsilon": [[0.8]], "sigma": [[1.0]]}},
                {"type": "lennard_jones", "parameters": {"epsilon": [[0.2]], "sigma": [[1.0]]}}
            ]
        }
        system = self.trajectory[0]
        system.interaction = Interaction(model)
        system = change_species(system, 'F')
        self.assertAlmostEqual(system.potential_energy(per_particle=True), -3.8079776291909284)
        self.assertAlmostEqual(system.interaction.forces[0, 0], 65.3083639139375)
        # This is not possible yet
        # system.compute_interaction('gradw')
        # self.assertAlmostEqual(system.interaction.gradw[0, 0], -71505.20734345658)
        system.compute_interaction('hessian')
        self.assertAlmostEqual(system.interaction.hessian[0, 0, 0, 0], 963.3891474196369)

    def test_neighbors(self):
        from atooms.trajectory.decorators import change_species
        system = self.trajectory[0]
        system.interaction = Interaction(self.model)
        system.interaction.neighbor_list = VerletList()
        system = change_species(system, 'F')
        self.assertAlmostEqual(system.potential_energy(per_particle=True), -3.8079776291909284)
        self.assertAlmostEqual(system.interaction.forces[0, 0], 65.3083639139375)
        # This is not implemented yet
        # system.compute_interaction('gradw')
        # self.assertAlmostEqual(system.interaction.gradw[0, 0], -71505.20734345658)
        system.compute_interaction('hessian')
        self.assertAlmostEqual(system.interaction.hessian[0, 0, 0, 0], 963.3891474196369)
        system.compute_interaction('energies')
        self.assertAlmostEqual(system.potential_energy(cache=True), sum(system.interaction.energies) / 2)

    def test_neighbors_update(self):
        system = self.trajectory[0]
        system.interaction = Interaction(self.model)
        system.interaction.neighbor_list = VerletList()
        system.species_layout = 'F'
        self.assertAlmostEqual(system.potential_energy(per_particle=True), -3.8079776291909284)
        nmax = system.interaction.neighbor_list.neighbors.shape[0]
        system.density *= 1.5
        system.potential_energy(per_particle=True)
        self.assertTrue(system.interaction.neighbor_list.neighbors.shape[0], nmax)

    def test_neighbors_lattice(self):
        import random
        import numpy
        random.seed(1)
        # This will be a perfect simple cubic
        s = System(N=125)
        s.density = 1.0
        s.species_layout = 'F'
        nl_sann = NeighborList(2.0, method='sann')
        nl_sann.compute(s.dump('box'), s.dump('pos', order='F'), s.dump('species'))
        nl = NeighborList([[1.1]], full=True)
        nl.compute(s.dump('box'), s.dump('pos', order='F'), s.dump('species'))
        i = 0
        # for j in range(nl_sann.number_neighbors[i]):
        #     print(nl_sann.distances[j, i]**0.5, nl_sann.neighbors[j, i])
        # print()
        # for j in range(nl.number_neighbors[i]):
        #     print(nl.distances[j, i]**0.5, nl.neighbors[j, i])
        nn = sorted(nl.neighbors[0:nl.number_neighbors[i], i])
        nn_sann = sorted(nl_sann.neighbors[0:nl_sann.number_neighbors[i], i])
        self.assertEqual(nn, nn_sann)
        self.assertTrue(numpy.all(nl.distances[0:nl_sann.number_neighbors[i]] < 1.1))
        self.assertEqual(len(nn), 6)

    def test_neighbors_lattice_binning(self):
        self.skipTest('timings are unreliable')
        import random
        import numpy
        for d, N in [(2, 10000), (3, 20000)]:
            random.seed(1)
            s = System(N=N, d=d)
            s.composition = {'A': N//2, 'B': N//2}
            s.density = 1.0
            s.species_layout = 'F'
            s.interaction = Interaction(self.model)
            s.interaction.neighbor_list = VerletList(binning=True)
            from atooms.core.utils import Timer
            with Timer(output=None) as timer1:
                epot_0 = s.potential_energy(per_particle=True)
            s.interaction.neighbor_list = VerletList(binning=False)
            with Timer(output=None) as timer2:
                epot_1 = s.potential_energy(per_particle=True)
            self.assertLess(timer1.wall_time, timer2.wall_time)
            self.assertAlmostEqual(epot_0, epot_1, places=7)
            # print(timer1.wall_time, timer2.wall_time)

    def test_sort(self):
        src = """
  SUBROUTINE sort(neighbor, distance, sortneighbor, distancesorted)
    ! Parameters
    INTEGER(8), INTENT(inout) :: neighbor(:)
    REAL(8), INTENT(inout) :: distance(:)
    INTEGER(8), INTENT(out) :: sortneighbor(size(neighbor))
    REAL(8), INTENT(out) :: distancesorted(size(neighbor))
    ! Variables
    INTEGER(8) :: i, imin, j, n_tmp
    REAL(8)    :: d_tmp
    ! Computation
    DO i=1,size(neighbor)
      imin = i
      DO j=i+1,size(neighbor)
        IF (distance(j) < distance(imin)) THEN
          imin = j
        END IF
        d_tmp = distance(i)
        n_tmp = neighbor(i)
        distance(i) = distance(imin)
        neighbor(i) = neighbor(imin)
        distance(imin) = d_tmp
        neighbor(imin) = n_tmp
        distancesorted(i) = distance(i)
        sortneighbor(i) = neighbor(i)
     END DO
     if (i == size(neighbor)) sortneighbor(i) = neighbor(i)
     if (i == size(neighbor)) distancesorted(i) = distance(i)
     print*, 'sort', i, imin, neighbor(i), sortneighbor(i)
    END DO
  END SUBROUTINE   
"""
        src = """
 subroutine insertion_sort(n, x, ns, xs)
    implicit none
    integer(8), intent(inout) :: n(:)
    real(8), intent(inout) :: x(:)
    integer(8), intent(out) :: ns(size(n))
    real(8), intent(out) :: xs(size(n))
    integer(8) :: i,j
    integer(8) :: temp
    real(8) :: temp_x

    ns = n
    xs = x
    do i=2,size(ns)
       temp=ns(i)
       temp_x=xs(i)
       do j=i-1,1,-1
          if (xs(j) <= temp_x) exit
          ns(j+1)=ns(j)
          xs(j+1)=xs(j)
       enddo
       ns(j+1)=temp
       xs(j+1)=temp_x
    enddo
  end subroutine insertion_sort
"""
        import f2py_jit
        import numpy
        f90 = f2py_jit.jit(src)
        n = numpy.array([1, 2, 3, 4])
        x = numpy.array([2.0, 3.0, 1.0, 1.1])
        ns, xs = f90.insertion_sort(n, x)
        # print(ns, xs)

    @unittest.skip('broken test')
    def test_neighbors(self):
        from atooms.trajectory.decorators import change_species
        rcut = [[1.5]]
        system = self.trajectory[0]
        system.neighbor_list = NeighborList(rcut)
        system.species_layout = 'F'
        system.compute_neighbor_list()
        self.assertEqual(list(system.particle[0].neighbors), [17, 32, 52, 91, 109, 112, 121, 140, 149, 162, 181, 198, 231, 247, 256])

    def test_parallel(self):
        from atooms.trajectory import Trajectory, change_species
        from atooms.simulation import Simulation, Scheduler, write_config, write_thermo

        def main(parallel, neighbor):
            fileinp = os.path.join(os.path.dirname(__file__), '../data/lj_N1000_rho1.0.xyz')
            self.trajectory.close()
            self.trajectory = Trajectory(fileinp)
            system = self.trajectory[-1]
            system.species_layout = 'F'
            system.interaction = Interaction(self.model, parallel=parallel)
            if neighbor:
                system.interaction.neighbor_list = VerletList(skin=0.3, parallel=parallel, update='periodic', update_period=1)
            system.compute_interaction()
            return system.potential_energy(per_particle=True, cache=True)

        # Note that threads should be set from outside
        u_p = main(True, True)
        u_s = main(False, True)
        self.assertTrue(abs(u_p - u_s) < 1e-10)
        u_p = main(True, False)
        u_s = main(False, False)
        self.assertTrue(abs(u_p - u_s) < 1e-10)

    def test_polydisperse(self):
        fileinp = os.path.join(os.path.dirname(__file__), '../data/lj_poly_N250.xyz')
        trajectory = Trajectory(fileinp)
        model = {
            "cutoff": [{
                "type": "cut_shift", "parameters": {"rcut": [[1.25]]}
            }],
            "potential": [{
                "type": "lennard_jones",
                "parameters": {"sigma": [[1.0]],
                               "epsilon": [[1.0]]}
            }],
        }
        system = trajectory[0]
        system.species_layout = 'F'
        system.interaction = Interaction(model)
        self.assertAlmostEqual(system.potential_energy(per_particle=True), 0.2677241464681521)
        self.assertAlmostEqual(system.interaction.forces[0, 0], 0.826650238231295)
        trajectory.close()

    def test_cutoff(self):
        import f2py_jit
        for cut in ["cut", "cut_shift", "cut_shift_linear",
                    "linear_cut_shift", "quadratic_cut_shift",
                    "cut_shift_quadratic", "cubic_spline"]:
            self.model['cutoff'][0]['type'] = cut
            self.model['cutoff'][0]['parameters']['rcut'][0] = 1.25
            interaction = Interaction(self.model)
            f90 = f2py_jit.import_module(interaction._uid[0])
            # The logical is returned as integer
            self.assertEqual(f90.cutoff.is_zero(1, 1, (1.25-0.01)**2), 0)
            self.assertEqual(f90.cutoff.is_zero(1, 1, (1.25+0.01)**2), 1)

    def test_potential_derivatives(self):
        from atooms.backends.f90.helpers import _check_potential_derivatives as _check
        self.assertTrue(_check('gaussian', [1.0, 1.0], 1.0))
        self.assertTrue(_check('harmonic', [1.0, 1.0], 1.0))
        self.assertTrue(_check('yukawa', [1.0, 1.0, 0.0], 1.0))
        self.assertTrue(_check('lennard_jones', [1.0, 1.0], 1.0))
        self.assertTrue(_check('inverse_power', [12, 1.0, 1.0], 1.0))
        self.assertTrue(_check('sum_inverse_power', [[12], [1.0], [1.0]], 1.0))

    def test_interaction_field_total(self):
        import numpy
        from atooms.backends.f90 import Interaction as InteractionParticle
        from atooms.system.interaction import InteractionField, Interaction

        system = self.trajectory[0]
        system.species_layout = 'F'

        def gravitational_field(pos, mass, g=9.81):
            u = mass[:] * g * pos[-1, :]
            grad = numpy.zeros_like(pos)
            grad[-1, :] = mass[:] * g
            return u, grad, None
        system.interaction = Interaction()
        system.interaction.add(InteractionParticle(self.model))
        system.interaction.add(InteractionField(gravitational_field,
                                                variables={'pos': 'particle.position',
                                                           'mass': 'particle.mass'}))
        system.compute_interaction('forces')

    def test_helpers(self):
        """Test that unrolling works for d=2 and d=3"""
        import sys
        from atooms.core.utils import Timer

        if sys.version_info.major == 3 and sys.version_info.minor >= 6:
            self.skipTest('manual loop unrolling is ineffective with 3.6')

        # Check 3d
        system = self.trajectory[0]
        system.species_layout = 'F'
        system.interaction = Interaction(self.model)
        with Timer(output=None) as timer1:
            u1 = system.compute_interaction('energy')
        system.interaction = Interaction(self.model, dimensions=3)
        with Timer(output=None) as timer2:
            u2 = system.compute_interaction('energy')
        self.assertAlmostEqual(u1, u2)
        # print(timer2.wall_time, timer1.wall_time)
        self.assertLess(timer2.wall_time, timer1.wall_time)

        # Check 2d (could be refactored)
        system.cell.side = system.cell.side[:2]
        # system.cell.origin = system.cell.origin[:2]
        for p in system.particle:
            p.position = p.position[:2]
            p.velocity = p.velocity[:2]
        system.interaction = Interaction(self.model)
        with Timer(output=None) as timer1:
            u1 = system.compute_interaction('energy')
        system.interaction = Interaction(self.model, dimensions=2)
        with Timer(output=None) as timer2:
            u2 = system.compute_interaction('energy')
        self.assertAlmostEqual(u1, u2)
        # print(timer2.wall_time, timer1.wall_time)

    def tearDown(self):
        self.trajectory.close()


if __name__ == '__main__':
    unittest.main()



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

"""
Folder based trajectory.

This is a base class for trajectories that stores configurations as
individual files in a directory. The `dirname` property stores the
directory path; an additional `files` property stores the list of
files sorted by step index.

It supports compressed archives. In this case the input filename is
the path to the archive, which will be decompressed in a temporary
folder.
"""

import os
import glob
import re
import tarfile
import tempfile

from atooms.core.utils import rmd
from atooms.trajectory.base import TrajectoryBase


# Helper functions

def init_folder(filename, file_pattern='*', step_pattern=r'(\d*)'):
    """
    Initial setup in read mode.

    Filename can be a folder, a compressed tar file, a simple file.
    """
    if os.path.isdir(filename):
        # Configurations are stored in a folder
        dirname = filename
        archive = False
        files = glob.glob(os.path.join(dirname, file_pattern))
    else:
        # Trajectory is packed as a compressed tar file.
        # If so, configurations will be extracted inplace and deleted at the end.
        with tarfile.open(filename) as th:
            dirname = tempfile.mkdtemp()

            def is_within_directory(directory, target):

                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)

                prefix = os.path.commonprefix([abs_directory, abs_target])

                return prefix == abs_directory

            def safe_extract(tar, path=".", members=None, *, numeric_owner=False):

                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")

                tar.extractall(path, members, numeric_owner=numeric_owner)

            safe_extract(th, path=dirname)
            files = [os.path.join(dirname, f.name) for f in th.getmembers()]
        archive = True

    files, steps = _get_file_steps(files, step_pattern)
    return dirname, archive, files, steps

def _get_step(fileinp, step_pattern):
    # Make sure we only test the basename (avoid metching
    # patterns in directory path)
    s = re.search(step_pattern, os.path.basename(fileinp))
    if s:
        step = int(s.group(1))
        return step
    else:
        raise ValueError('Could not find step')

def _get_file_steps(files, step_pattern):
    """
    Return a list of tuples (file, step). The step is extracted from
    the file path using `regexp`, which must contain one group for the
    step.
    """
    file_steps = []
    for i, f in enumerate(files):
        if os.path.isdir(f):
            continue
        try:
            step = _get_step(f, step_pattern)
        except ValueError:
            step = i+1
        file_steps.append((f, step))
    file_steps.sort(key=lambda a: a[1])
    return [a[0] for a in file_steps], [a[1] for a in file_steps]


# Classes

class TrajectoryFolder(TrajectoryBase):

    """Folder based trajectory."""

    def __init__(self, filename, mode='r', file_pattern='*', step_pattern=r'(\d*)'):
        TrajectoryBase.__init__(self, filename.rstrip('/'), mode)
        if mode == 'r':
            output = init_folder(filename, file_pattern, step_pattern)
            self.dirname, self.archive, self.files, self.steps = output
        else:
            self.dirname = filename
            self.archive = False
            self.files, self.steps = [], []

    def close(self):
        if self.archive:
            rmd(self.dirname)


class Foldered(TrajectoryFolder):

    """Transform a file-based trajectory into folder-based one. Read-only."""

    def __init__(self, filename, mode='r', cls=None, file_pattern='*', step_pattern=r'(\d*)'):
        if mode != 'r':
            raise ValueError('Not ready for write mode')
        TrajectoryFolder.__init__(self, filename, mode)
        self._cls = cls
        if self.mode == 'r':
            self.dirname, self.archive, self.files, self.steps = init_folder(self.filename)

    def read_timestep(self):
        from atooms.trajectory import Trajectory
        with Trajectory(self.files[0], fmt=self._cls) as th:
            return th.read_timestep()

    def read_system(self, frame):
        from atooms.trajectory import Trajectory
        with Trajectory(self.files[frame], fmt=self._cls) as th:
            # We must use read(), instead of read_system(), to
            # initialize the trajectory properly
            return th.read(0)

    def close(self):
        if self.archive:
            rmd(self.dirname)



# This file is part of atooms
# Copyright 2010-2024, Daniele Coslovich

import warnings
import numpy
import re
import logging
from copy import copy
from .simple import TrajectorySimpleXYZ
from .base import TrajectoryBase
from .utils import gopen, _PossiblyZippedFile
from atooms.core.utils import tipify
from atooms.system.particle import Particle, distinct_species
from atooms.system.cell import Cell
from atooms.system import System

log = logging.getLogger(__name__)

# Lock to avoid race conditions on numpy array formatting
_numpy_fmt_lock = False

# Format callbacks

def _update_radius(particle, data, meta):
    particle.radius = float(data[0])
    return data[1:]

def _update_diameter(particle, data, meta):
    particle.radius = float(data[0]) / 2
    return data[1:]

def _update_tag(particle, data, meta):
    # Kept for backward compatibility. We should rather rely on
    # tipified properties.
    particle.tag = data[0:]
    return data[1:]

def _update_species(particle, data, meta):
    particle.species = data[0]
    return data[1:]

def _update_position(particle, data, meta):
    ndim = meta['ndim']
    # It is crucial to assing position, not to use the slice!
    # Otherwise we get a reference, not a copy.
    particle.position = numpy.array(data[0:ndim], dtype=float)
    return data[ndim:]

def _update_position_unfolded(particle, data, meta):
    from atooms.system.particle import _periodic_vector_unfolded
    ndim = meta['ndim']
    r = numpy.array(data[0:ndim], dtype=float)
    particle.position_unfolded = r
    if 'position' not in meta['columns']:
        particle.position = _periodic_vector_unfolded(r, meta['cell'])
    return data[ndim:]

def _update_velocity(particle, data, meta):
    ndim = meta['ndim']
    particle.velocity = numpy.array(data[0:ndim], dtype=float)
    return data[ndim:]

def _update_neighbors_consume(particle, data, meta):
    # Consume all entries in data
    particle.neighbors = numpy.array(data, dtype=int)
    return []

def _update_neighbors(particle, data, meta):
    # Extract comma separated entries in the first element of data
    if len(data) > 0 and ',' in data[0]:
        if data[0] == ',':
            particle.neighbors = numpy.array([], dtype=int)
        else:
            particle.neighbors = numpy.array(data[0].split(','), dtype=int)
        return data[1:]
    else:
        particle.neighbors = numpy.array(data, dtype=int)
        return []

def _optimize_arrays(variables):
    """
    Optimize a list of particles properties (`variables`) by
    collapsing successive instances of array elements.

    Example: ['id', 'x[0]', 'x[1]', 'x[2]'] -> ['id', 'x']

    > print(scalars_to_arrays(['pos[0]', 'pos[1]', 'pos[2]']))
    > print(scalars_to_arrays(['pos[0]', 'pos[1]', 'vel[0]', 'vel[1]']))
    > print(scalars_to_arrays(['pos[0]', 'pos[1]', 'pos[2]', 'hello']))
    > print(scalars_to_arrays(['hello', 'pos[0]', 'pos[1]', 'pos[2]']))
    """
    import re

    new_variables = []
    last_var, last_idx = None, None
    for variable in variables:
        # This will match array expressions, like position[0] or particle.position[0]
        all_matches = re.findall(r'([\._\w]+)\[(\d+)\]', variable)

        # This must be a single couple (variable, index) or nothing
        if len(all_matches) == 1:
            match = all_matches[0]
        elif len(all_matches) == 0:
            match = []
        else:
            raise ValueError('problem with variable {}'.format(variable))

        if len(match) == 2:
            # This is an array element
            # Extract variable name and array index
            var, idx = match[0], int(match[1])
            if last_var != var and last_idx is not None:
                # We have two consecutive different arrays
                new_variables.append(last_var)
            last_var = var
            last_idx = idx
        elif len(match) == 0:
            var = variable
            if last_var != var and last_idx is not None:
                # We just finished reading an array variable
                # We store it and also the next
                new_variables.append(last_var)
            new_variables.append(var)
            last_var = var
            last_idx = None
        else:
            raise ValueError('problem with variable {}'.format(variable))

    if last_idx is not None:
        # We get here if the last was an array element
        new_variables.append(last_var)
    return new_variables


class TrajectoryXYZ(TrajectoryBase):

    """
    XYZ format with metadata support (https://en.wikipedia.org/wiki/XYZ_file_format)

    It is implemented using a lightweight indexing approach.
    """

    suffix = 'xyz'
    callback_read = {'particle.species': _update_species,
                     'particle.radius': _update_radius,
                     'particle.position': _update_position,
                     'particle.velocity': _update_velocity,
                     'particle.position_unfolded': _update_position_unfolded,
                     'position_unfolded': _update_position_unfolded,
                     'diameter': _update_diameter,
                     'name': _update_species,
                     'tag': _update_tag,
                     'neighbors': _update_neighbors,
                     'neighbors*': _update_neighbors_consume}

    def __init__(self, filename, mode='r', alias=None, fields=None):
        super(TrajectoryXYZ, self).__init__(filename, mode)
        self.variables = [
            'particle.species',
            'particle.position'
        ]
        if fields is not None:
            self.variables = fields
            warnings.warn('fields is deprecated, set trajectory.variables instead', FutureWarning)

        self.alias = {} if alias is None else alias
        # Internal variables
        self._cell = None
        # Trajectory file handle
        # self._file = gopen(self.filename, self.mode)
        self._file = _PossiblyZippedFile(self.filename, self.mode)

        if self.mode == 'r':
            # Internal index of lines via seek and tell.
            self._setup_index()
            # Initialize metadata and variables
            self.metadata = self._read_comment(0)
            if 'columns' in self.metadata:
                self.variables = self.metadata['columns']
            self._original_variables = copy(self.variables)
            self._active_read_callbacks = None

    def _setup_format(self):
        # This private method has been removed, but might have been used in client code.
        # If so, we raise an exception to at least avoid unexpected results.
        raise AttributeError('method _setup_fmt() was removed')

    def _setup_index(self):
        """Sample indexing via tell / seek"""
        self._index_frame = []
        self._index_header = []
        self._file.seek(0)
        while True:
            line = self._file.tell()
            data = self._file.readline()

            # We break if file is over or we found an empty line
            if not data:
                break

            # The first line contains the number of particles
            # If something goes wrong, this could be the last line
            # with the cell side (Lx,Ly,Lz). We do not support this anymore
            try:
                npart = int(data)
                self._index_header.append(line)
            except ValueError:
                break

            # Skip npart+1 lines, making sure we have read precisely
            # that number of lines
            data = self._file.readline()
            line = self._file.tell()
            for _ in range(npart):
                data = self._file.readline()
            # Store first line /after/ we have read the frame
            # making sure the last we read was not emtpy
            # Note that readline() returns an empty string on EOF
            if len(data) > 0:
                self._index_frame.append(line)
            else:
                raise IOError('malformed xyz file [%s]', self.filename)

        # Sanity tests
        assert len(self._index_frame) > 0, 'empty file {}'.format(self._file)
        assert len(self._index_header) > 0, 'empty file {}'.format(self._file)

    # Overwrite the variables setter to optimize the variables
    @TrajectoryBase.variables.setter
    def variables(self, value):
        from atooms.core.utils import canonicalize

        self._variables = canonicalize(value, self.thesaurus)
        self._variables = _optimize_arrays(self._variables)

    def read_steps(self):
        """Find steps list."""
        steps = []
        for frame in range(len(self._index_frame)):
            meta = self._read_comment(frame)
            try:
                steps.append(meta['step'])
            except KeyError:
                # If no step info is found, we add steps sequentially
                steps.append(frame+1)
        return steps

    def _read_comment(self, frame):
        """
        Internal xyz method to get header metadata from comment line of
        given `frame`.

        We assume metadata fmt is a space-separated sequence of comma
        separated entries such as:

        columns:id,x,y,z step:10
        columns=id,x,y,z step=10
        """
        # Go to line and skip Npart info
        self._file.seek(self._index_header[frame])
        npart = int(self._file.readline())
        data = self._file.readline()
        meta = {}
        meta['npart'] = npart

        # Read metadata line
        if not ('=' in data or ':' in data):
            # The comment line contains a list of unspecified fields
            # We add them to the dict using a sequential integer key
            for i, value in enumerate(data.split()):
                meta[i] = tipify(value)

        else:
            # The comment line contains self descriptive fields
            # Remove spaces around : or = and replace = by :
            data = re.sub(r'\s*[=:]\s*', ':', data)
            # Tolerate spaces between entries of vectors
            data = re.sub(r'\s*,\s*', ',', data)

            # Fill metadata dictionary
            for e in data.split():
                s = re.search(r'(\S+):(\S+)', e)
                if s is not None:
                    tag, data = s.group(1), s.group(2)
                    # Remove dangling commas
                    data = data.strip(',')
                    # If there are commas, this is a list, else a scalar.
                    # We convert the string to appropriate types
                    if ',' in data:
                        meta[tag] = [tipify(_) for _ in data.split(',')]
                    else:
                        meta[tag] = tipify(data)

        # Apply an alias dict to tags, e.g. to add step if Cfg was found instead
        for alias, tag in self.alias.items():
            try:
                meta[tag] = meta[alias]
            except KeyError:
                pass

        # Fix dimensions based on side of cell.
        # Fallback to ndim metadata or 3.
        try:
            if 'ndim' not in meta:
                meta['ndim'] = len(meta['cell'])
        except TypeError:
            meta['ndim'] = 1  # it is a scalar
        except KeyError:
            meta['ndim'] = 3  # default

        # Make sure columns is a list
        if 'columns' in meta:
            if not isinstance(meta['columns'], list):
                meta['columns'] = [meta['columns']]

        return meta

    def _setup_read_callbacks(self):
        """
        Return a list of callbacks to read the variables given in self.variables
        """
        # We only call this method once, when the we read the first sample
        # If self.variables change after setup, we ignore the change
        # Similarly, we assume that columns do not change
        if self._active_read_callbacks is not None:
            return self._active_read_callbacks

        # Define active callbacks before reading the sample. We
        # handle the case in which the user removed some variables,
        # and self.variables is a subset of variables_in_files
        callbacks = []
        test_particle = Particle()
        for key in self._original_variables:
            if key not in self.variables:
                # If the key is not the in original list of variables
                # the user has removed it from the list and we skip it
                nsafe = 20
                cbk = self.callback_read[key]
                consumed_entries = nsafe - len(cbk(test_particle, range(nsafe), self.metadata))
                # Trick. We instantiate dynamically a skip function
                # with the right number of consumed entries
                namespace = {}
                exec("""
def skip(p, data, meta):
    return data[{}:]
""".format(consumed_entries), namespace)
                callbacks.append(namespace['skip'])
            elif key in self.callback_read:
                # The key is associated to a explicit callback
                callbacks.append(self.callback_read[key])
            else:
                # We throw the field in an particle.
                # Trick. We instantiate dynamically a fallback function
                # to avoid adding `key` to the other callbacks' interface
                namespace = {}
                exec("""
from atooms.core.utils import tipify
def fallback(p, data, meta):
    p.__dict__['%s'] = tipify(data[0])
    return data[1:]
""" % key, namespace)
                callbacks.append(namespace['fallback'])

        self._active_read_callbacks = callbacks

    def read_system(self, frame):
        # Define actual list of callbacks
        self._setup_read_callbacks()
        callbacks = self._active_read_callbacks
        # Read metadata of this frame
        db = self._read_comment(frame)

        # Read frame now
        self._file.seek(self._index_frame[frame])
        particle = []
        for _ in range(db['npart']):
            p = Particle()
            # Note: we cannot optimize by shifting an index instead of
            # cropping lists all the time
            data = self._file.readline().split()
            for cbk in callbacks:
                data = cbk(p, data, db)
            particle.append(p)

        # Fix the masses.
        # We assume masses read from the header are sorted by species name.
        # The mass dbdata must be adjusted to the given frame.
        if 'mass' in db:
            if isinstance(db['mass'], list) or isinstance(db['mass'], tuple):
                species = distinct_species(particle)
                # We must have as many mass entries as species
                if len(species) != len(db['mass']):
                    raise ValueError('mass metadata issue %s, %s' %
                                     (species, db['mass']))
                db_species = {}
                for key, value in zip(species, db['mass']):
                    db_species[key] = value
                for p in particle:
                    p.mass = float(db_species[p.species])
            else:
                for p in particle:
                    p.mass = float(db['mass'])

        # Add cell info
        if 'cell' in db:
            cell = Cell(db['cell'])
        else:
            cell = None

        return System(particle, cell)

    def read_timestep(self):
        if 'dt' in self.metadata:
            return self.metadata['dt']
        elif 'timestep' in self.metadata:
            return self.metadata['timestep']
        else:
            return 1.0

    def _comment(self, step, system):
        # Variables in xyz files are assumed to be referred to particles
        # so we strip 'particle' from all column entries
        variables = [re.sub('particle.', '', var) for var in self.variables]
        # Concatenate metadata in comment line
        line = 'step:{} '.format(step)
        line += 'columns:{} '.format(','.join(variables))
        if self.timestep is not None:
            line += 'dt:{:g} '.format(self.timestep)
        if system.cell is not None:
            line += 'cell:{} '.format(','.join([str(x) for x in system.cell.side]))
        for entry in self.metadata:
            line += '{}:{} '.format(entry, self.metadata[entry])
        return line

    def write_system(self, system, step):
        self._file.write('%d\n' % len(system.particle))
        self._file.write(self._comment(step, system) + '\n')

        def array_str(arr):
            """Format numpy arrays or scalars"""
            # Passing a scalar will trigger an error (gotcha: even
            # when casting numpy array to list, the elements remain of
            # numpy type and this function gets called! (4% slowdown)
            _fmt_any = '{}'
            _fmt_float = '{:.' + str(self.precision) + 'f}'
            if arr.dtype.str[1] == 'f':  # this matches both float32 and float64
                _fmt = _fmt_float
            else:
                _fmt = _fmt_any
            if len(arr.shape) == 1:
                return ' '.join([_fmt.format(x) for x in arr])
            elif len(arr.shape) == 0:
                return _fmt.format(arr)
            else:
                raise ValueError('cannot write multi-dimensional array')

        # In numpy 2 we can only override repr() to get a reasonable efficiency.
        # But this means we must use a switch for non-array variables.
        def str_or_repr(x):
            if isinstance(x, numpy.ndarray):
                return array_str(x)
            else:
                return str(x)
            
        variables = [v.split('particle.')[-1] for v in self.variables]
        for i, p in enumerate(system.particle):
            p._index = i
            p._step = step
            line = []
            for v in variables:
                line.append(str_or_repr(getattr(p, v)))
            self._file.write(' '.join(line) + '\n')

    def close(self):
        self._file.close()


def _add_neighbors_to_system(system, offset):
    for p in system.particle:
        p.neighbors -= offset
    system.neighbors = [p.neighbors for p in system.particle]
    return system


class TrajectoryNeighbors(TrajectoryXYZ):

    """
    Neighbors trajectory format

    By default, for reading we assume an xyz file with space-separated
    integers indicating the particles indices. For writing, we use
    comma-separated entries (no space).
    """

    def __init__(self, filename, mode='r', fields=None, offset=1):
        super(TrajectoryNeighbors, self).__init__(filename, mode=mode,
                                                  alias={'time': 'step'})
        self._offset = offset
        self.thesaurus = {'neighbors': 'particle.neighbors',
                          'neighbors*': 'particle.neighbors*', }
        if mode == 'w':
            self.variables = ['neighbors'] if fields is None else fields
        if mode == 'r' and 'columns' not in self.metadata:
            self.variables = ['neighbors*']
        self.callback_read['particle.neighbors'] = _update_neighbors
        self.callback_read['particle.neighbors*'] = _update_neighbors_consume
        self.add_callback(_add_neighbors_to_system, self._offset)



import numpy
import f2py_jit

from atooms.system.interaction import InteractionBase
from .helpers import _merge_source, _normalize_path


def _check_polydisperse(species, radius):
    """Check whether system is polydisperse"""
    # Make sure all species are identical
    poly = False
    tolerance = 1e-15
    if numpy.all(species == species[0]):
        delta = abs(radius.min() - radius.max())
        if delta > tolerance * radius.mean():
            poly = True
    return poly


# Aliases for cutoffs and potentials (for backward compatibility)
# First entry of tuple is the alias, second one is the official name
aliases = [('linear_cut_shift', 'cut_shift_linear'),
           ('quadratic_cut_shift', 'cut_shift_quadratic'),
           ('harmonic_potential', 'harmonic'),
           ]

class Interaction(InteractionBase):

    """
    A fortran 90 backend to compute interactions between particles.

    Note: we expect particle species to be either integers or
    strings that can be cast to integers (corresponding either to
    `System.species_layout` equal to 'F' or 'C', but not to 'A').
    If you encounter a `ValueError`, convert the species layout with
    `system.species_layout = 'F'`.

    Under the hoods, the idea is to stitch together four fortran modules:

    1) a helpers module, which define common subroutines needed to
    compute several types of interaction terms

    2) a potential module, which defines subroutines to initialize and
    compute a potential between particles

    3) a cutoff module, which defines subroutines to initialize, cut
    off and smooth a potential around some distance

    4) an interaction module, which uses the above ones to compute the
    total interaction energy, forces, etc.

    The above modules are merged in a single source at run time and
    compiled using f2py-jit.
    """

    def __init__(self, model, neighbor_list=None,
                 interaction='interaction.f90', helpers='helpers.f90',
                 inline=True, inline_safe=False, debug=False,
                 parallel=False, search_dir=None, dimensions=None):
        """
        The interaction `model` is a dictionary with "potential" and
        "cutoff" keys. Each key is a list of dictionaries containing
        potential types and parameters, with the following layout:

        .. code-block:: python

           model = {
             "potential": [{
                 "type": "lennard_jones",
                 "parameters": {"epsilon": [[1.0]], "sigma": [[1.0]]}
             }],
             "cutoff": [{
                 "type": "cut_shift",
                 "parameters": {"rcut": [[2.5]]}
             }]
           }

        If multiple potentials are found in the list, the
        corresponding interaction terms are summed up.

        The `type` variables must match the names of the fortran
        modules defined by the `f90` backend.

        The dictionaries `parameters` must match the input arguments
        of the `init` subroutines of the selected fortran modules.

        The entries of the `parameters` dictionaries must be (nsp,
        nsp) arrays, where nsp is the number of chemical species
        defined by the model (in the above example, nsp=1).

        The `name` key can contain a string that defines the name of
        the model.
        """
        InteractionBase.__init__(self)
        self.neighbor_list = neighbor_list
        self.order = 'F'
        self.observable.append('gradw')
        # TODO: relax ids using int64 to avoid changes of type
        self.variables = {'box': 'cell.side:float64',
                          'pos': 'particle.position:float64',
                          'ids': 'particle.species:int32',
                          'rad': 'particle.radius:float64'}
        self.model = model
        # Curiously, it is not convenient to manually unroll
        # the loops with dimensions = 2 (see test_helpers)
        if dimensions == 3:
            helpers = f'helpers_{dimensions}d.f90'

        # Cache for polydisperse system
        self._polydisperse = None
        # Cache for species layout (we tolerate C layout by adding +1 later)
        self._species_layout = 'F'

        # At this stage we expect a model dictionary
        # Normalize paths
        for entry in model["potential"] + model["cutoff"]:
            for alias in aliases:
                if entry["type"] == alias[0]:
                    entry["type"] = alias[1]
                    break
            if "path" not in entry:
                entry["path"] = _normalize_path(entry.get("type"), search_dir)

        # Loop over individual potentials and build interaction modules
        self._uid = []
        for _potential, _cutoff in zip(model['potential'], model['cutoff']):
            potential = _potential.get('path')
            potential_parameters = _potential.get('parameters')
            cutoff = _cutoff.get('path')
            cutoff_parameters = _cutoff.get('parameters')
            potential = _normalize_path(potential, search_dir)
            cutoff = _normalize_path(cutoff, search_dir)

            # Number of bodies. It cannot be inferred reliably from the shape of the
            # parameters arrays. Either check an explicit list of 2-body, 3-body, etc
            # potentials (hence this information is local in the f90 backend) or
            # expect a hint in the model dictionary, like bodies = 3.
            # We remove the check for now, because it breaks some potentials.
            # if n_bodies == 3:
            #     interaction = 'interaction_three_body.f90'

            # Merge all sources into a unique source blob
            source = _merge_source(helpers, potential, cutoff, interaction)

            # Inline subroutines
            if inline:
                from f2py_jit.finline import inline_source
                # TODO: depending on f2py-jit version we can inline compute and smooth as well but this should be checked for bacward compatibility
                if inline_safe:
                    source = inline_source(source, ignore='compute,smooth,adjust,forces')
                elif inline:
                    source = inline_source(source, ignore='adjust,forces')

            # Compile and bundle the module with f2py
            args = ''
            opt_args = '-ffree-form -ffree-line-length-none'
            if debug:
                opt_args += ' -O3 -pg -fbounds-check'
            else:
                opt_args += ' -O3 -ffast-math'
            if parallel:
                opt_args += ' -fopenmp'
                args += ' -lgomp'
            extra_args = '--opt="{}" {}'.format(opt_args, args)

            # Build a unique module for each potential
            # Every model with its own parameter combination corresponds to a unique module
            # and can be safely reused (up to changes in interaction / helpers)
            uid = f2py_jit.build_module(source,
                                        metadata={"interaction": interaction,
                                                  "helpers": helpers,
                                                  "parallel": parallel,
                                                  "potential": [potential, potential_parameters],
                                                  "cutoff": [cutoff, cutoff_parameters]},
                                        extra_args=extra_args)
            # Setup potential and cutoff parameters
            _interaction = f2py_jit.import_module(uid)
            _interaction.potential.init(**potential_parameters)
            _interaction.cutoff.init(**cutoff_parameters)

            # Store module name (better not store the module itself, else we cannot deepcopy)
            self._uid.append(uid)

    def __str__(self):
        txt = 'interaction: f90 backend\n'
        if isinstance(self.model, dict):
            import pprint
            txt += 'model: ' + pprint.pformat(self.model) + '\n'
        else:
            txt += 'model: ' + self.model + '\n'
        if self.neighbor_list:
            txt += 'neighbor list: ' + str(self.neighbor_list)
        return txt

    def compute(self, observable, box, pos, ids, rad):
        """
        Compute `observable` from this interaction
        """
        # Check if system is polydisperse (cached)
        if self._polydisperse is None:
            self._polydisperse = _check_polydisperse(ids, rad)
        if self.neighbor_list is None:
            self._compute(observable, box, pos, ids, rad)
        else:
            self._compute_with_neighbors(observable, box, pos, ids, rad)

    def _initialize(self, observable, pos, ids):
        # We initialize observables at the first evaluation
        # or reinitialize them if the number of particles changed
        init = False
        if observable in ['forces', 'energy', None]:
            if self.forces is None or self.forces.shape != pos.shape:
                self.forces = numpy.zeros_like(pos, order='F')
                init = True
            # We set these variables as 0-dim array to modify them inplace
            # We cast them as float at the end of compute()
            self.energy, self.virial = numpy.array(0.0), numpy.array(0.0)
        elif observable == 'energies':
            if not hasattr(self, 'energies') or self.energies.size != pos.shape[1]:
                self.energies = numpy.zeros(pos.shape[1])
                init = True
        elif observable == 'gradw':
            if not hasattr(self, 'gradw') or self.gradw.shape != pos.shape:
                self.gradw = numpy.zeros_like(pos, order='F')
                init = True
        elif observable == 'hessian' or self.hessian:
            ndim, N = pos.shape
            if self.hessian is None or self.hessian.shape != (ndim, N, ndim, N):
                self.hessian = numpy.ndarray((ndim, N, ndim, N), order='F')
                init = True

        # If we initialized arrays, check species layout
        # At this stage, we expect species as numpy.int32
        if init:
            if numpy.min(ids) == 0:
                self._species_layout = 'C'

    def _compute(self, observable, box, pos, ids, rad):
        # Initialize variables
        self._initialize(observable, pos, ids)
        # Handle species layout
        if self._species_layout == 'C':
            ids = ids + 1

        # We set the variables to zero for the first interaction
        zero = True
        for uid in self._uid:
            # Load appropriate module according to polydispersity
            _interaction = f2py_jit.import_module(uid)
            if not self._polydisperse:
                f90 = _interaction.interaction
            else:
                f90 = _interaction.interaction_polydisperse

            # Compute the interaction cumulating terms
            if observable in ['forces', 'energy', None]:
                f90.forces(zero, box, pos, ids, rad, self.forces, self.energy, self.virial)
            elif observable == 'energies':
                f90.energies(zero, box, pos, ids, rad, self.energies)
            elif observable == 'gradw':
                assert len(self._uid) == 1, 'cannot compute gradw with multiple potentials'
                f90.gradw(zero, box, pos, ids, rad, self.gradw)
            elif observable == 'hessian':
                f90.hessian(zero, box, pos, ids, rad, self.hessian)

            # If there is more than interaction we cumulate them
            # so we do not zero the variables anymore
            zero = False

        if observable in ['forces', 'energy', None]:
            self.energy, self.virial = float(self.energy), float(self.virial)

    def _compute_with_neighbors(self, observable, box, pos, ids, rad):
        # Initialize variables
        self._initialize(observable, pos, ids)
        # Handle species layout
        if self._species_layout == 'C':
            ids = ids + 1

        # We set the variables to zero for the first interaction
        zero = True
        for uid in self._uid:
            _interaction = f2py_jit.import_module(uid)
            if not self._polydisperse:
                f90 = _interaction.interaction_neighbors
                cutoff = _interaction.cutoff.rcut_
            else:
                f90 = _interaction.interaction_polydisperse_neighbors
                # We set the cutoff to the largest particles: the worst
                # case scenario of
                #
                # rcut(i,j) = rcut * (radius(i) + radius(j))
                #
                # is when radius(i) = radius(j) = max(radius).
                #
                # This is a bit inefficient but keeps the code more
                # general and it works also with linked cells
                largest_radius = numpy.max(rad)
                cutoff = _interaction.cutoff.rcut_ * (largest_radius*2)
            if observable in ['forces', 'energy', None]:
                self.neighbor_list.adjust(box, pos, cutoff)
                self.neighbor_list.compute(box, pos, ids)
                f90.forces(zero, box, pos, ids, rad,
                           self.neighbor_list.neighbors,
                           self.neighbor_list.number_of_neighbors,
                           self.forces, self.energy, self.virial)
            elif observable == 'energies':
                f90.energies(zero, box, pos, ids, rad,
                             self.neighbor_list.neighbors,
                             self.neighbor_list.number_of_neighbors,
                             self.energies)
            elif observable == 'gradw':
                raise ValueError('gradw not implemented with neighbors')
            elif observable == 'hessian':
                f90.hessian(zero, box, pos, ids, rad,
                            self.neighbor_list.neighbors,
                            self.neighbor_list.number_of_neighbors,
                            self.hessian)
            zero = False

        if observable in ['forces', 'energy', None]:
            self.energy, self.virial = float(self.energy), float(self.virial)
